#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
from __future__ import division

import math

import pcbnew
import FootprintWizardBase
import PadArray as PA


class circular_pad_array_wizard(FootprintWizardBase.FootprintWizard):

    def GetName(self):
        return "Circular Pad Array"

    def GetDescription(self):
        return "Circular array of pads"

    def GenerateParameterList(self):

        self.AddParam("Pads", "count", self.uInteger, 6, min_value=1, designator='n')
        self.AddParam("Pads", "center diameter", self.uMM, 5, min_value=0, designator='r', hint="Centre distance between pads")
        self.AddParam("Pads", "diameter", self.uMM, 1.5)
        self.AddParam("Pads", "drill", self.uMM, 0.8)
        self.AddParam("Pads", "angle", self.uDegrees, 0, designator='a')
        self.AddParam("Pads", "rectangle", self.uBool, False)
        self.AddParam("Pads", "pad 1 rectangle", self.uBool, False)

        self.AddParam("Pad rotation", "pad rotation", self.uBool, False, designator='r')
        self.AddParam("Pad rotation", "pad angle offset", self.uDegrees, 0, designator='o')

        self.AddParam("Numbering", "initial", self.uInteger, 1, min_value=1)
        #self.AddParam("Numbering", "increment", self.uInteger, 1, min_value=1)
        self.AddParam("Numbering", "clockwise", self.uBool, True)

        self.AddParam("Outline", "diameter", self.uMM, 7, designator='D')
        self.AddParam("Outline", "margin", self.uMM, 0.25, min_value=0.2)

    def CheckParameters(self):

        pads = self.parameters['Pads']
        numbering = self.parameters['Numbering']
        outline = self.parameters['Outline']
        padRotation = self.parameters['Pad rotation']

        # Check that pads do not overlap
        pad_dia = pcbnew.ToMM(pads['diameter'])
        centres = pcbnew.ToMM(pads['center diameter'])
        n_pads = pads['count']

        self.CheckParam('Pads','diameter',max_value=centres*math.pi/n_pads,info="Pads overlap")

        # Check that the pads fit inside the outline
        d_min = pad_dia + centres

        self.CheckParam("Outline","diameter",min_value=d_min, info="Outline diameter is too small")


    def GetValue(self):
        pins = self.parameters["Pads"]["count"]
        return "CPA_%d" % pins

    def BuildThisFootprint(self):

        pads = self.parameters['Pads']
        numbering = self.parameters['Numbering']
        outline = self.parameters['Outline']
        padRotation = self.parameters['Pad rotation']

        pad_size = pads['diameter']
        pad_shape = pcbnew.PAD_SHAPE_RECT if pads["rectangle"] else pcbnew.PAD_SHAPE_OVAL

        pad = PA.PadMaker(self.module).THPad(pads['diameter'], pads['diameter'], pads['drill'], shape=pad_shape)

        array = PA.PadCircleArray(
            pad, pads['count'], pads['center diameter'] / 2,
            angle_offset=pads["angle"],
            centre=pcbnew.VECTOR2I(0, 0),
            clockwise=numbering["clockwise"],
            padRotationEnable= padRotation["pad rotation"],
            padRotationOffset = padRotation["pad angle offset"])

        if pads["pad 1 rectangle"]:
            firstPad = PA.PadMaker(self.module).THPad(pads['diameter'], pads['diameter'], pads['drill'], shape=pcbnew.PAD_SHAPE_RECT)
            array.SetFirstPadType(firstPad)

        array.SetFirstPadInArray(numbering["initial"])

        array.AddPadsToModule(self.draw)

        # Draw the outline
        body_radius = outline['diameter'] / 2
        self.draw.SetLayer(pcbnew.F_Fab)
        self.draw.SetLineThickness( pcbnew.FromMM( 0.1 ) ) #Default per KLC F5.2 as of 12/2018
        self.draw.Circle(0, 0, body_radius)

        #silkscreen
        body_radius += pcbnew.FromMM(0.12)
        self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) ) #Default per KLC F5.1 as of 12/2018
        self.draw.SetLayer(pcbnew.F_SilkS)
        self.draw.Circle(0, 0, body_radius)

        # courtyard
        self.draw.SetLayer(pcbnew.F_CrtYd)
        self.draw.SetLineThickness(pcbnew.FromMM(0.05))
        self.draw.Circle(0, 0, body_radius + outline['margin'])

        # Text (reference and value) size
        text_size = self.GetTextSize()  # IPC nominal
        thickness = self.GetTextThickness()
        textposy = body_radius + self.draw.GetLineThickness()/2 + self.GetTextSize()/2 + thickness + + outline['margin']
        self.draw.Value( 0, textposy, text_size )
        self.draw.Reference( 0, -textposy, text_size )

        # Add a extra text (${REFERENCE}) on the F_Fab layer
        extra_text = pcbnew.PCB_TEXT( self.module )
        extra_text.SetLayer( pcbnew.F_Fab )
        extra_text.SetPosition( pcbnew.VECTOR2I( 0, 0) )
        extra_text.SetTextSize( pcbnew.VECTOR2I( text_size, text_size ) )
        extra_text.SetText( "${REFERENCE}" )
        self.module.Add( extra_text )

circular_pad_array_wizard().register()



#
# Example python script to generate a BOM from a KiCad generic netlist
# The KiCad generic xml netlist is expected to be encoded UTF-8
#
# Example: Sorted and Grouped HTML BOM with advanced grouping
#

"""
    @package
    Output: HTML
    Grouped By: Value, Part, Footprint, Tolerance, Manufacturer, Voltage, DNP
    Sorted By: Ref
    Fields: Ref, Qnty, Value, Part, Footprint, Description, Vendor, DNP

    Command line:
    python "pathToFile/bom_with_advanced_grouping.py" "%I" "%O.html"
"""

from __future__ import print_function

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import kicad_utils
import sys


# Start with a basic html template
html = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
    <head>
        <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
        <title>KiCad BOM Example 5</title>
    </head>
    <body>
    <h1><!--SOURCE--></h1>
    <p><!--DATE--></p>
    <p><!--TOOL--></p>
    <p><!--COMPCOUNT--></p>
    <table>
    <!--TABLEROW-->
    </table>
    </body>
</html>
    """

def myEqu(self, other):
    """myEqu is a more advanced equivalence function for components which is
    used by component grouping. Normal operation is to group components based
    on their Value and Footprint.

    In this example of a more advanced equivalency operator we also compare the
    custom fields Voltage, Tolerance and Manufacturer as well as the assigned
    footprint. If these fields are not used in some parts they will simply be
    ignored (they will match as both will be empty strings).

    """
    result = True
    if self.getValue() != other.getValue():
        result = False
    elif self.getPartName() != other.getPartName():
        result = False
    elif self.getFootprint() != other.getFootprint():
        result = False
    elif self.getField("Tolerance") != other.getField("Tolerance"):
        result = False
    elif self.getField("Manufacturer") != other.getField("Manufacturer"):
        result = False
    elif self.getField("Voltage") != other.getField("Voltage"):
        result = False
    elif self.getDNP() != other.getDNP():
        result = False

    return result

# Override the component equivalence operator - it is important to do this
# before loading the netlist, otherwise all components will have the original
# equivalency operator.
kicad_netlist_reader.comp.__eq__ = myEqu

# Generate an instance of a generic netlist, and load the netlist tree from
# <file>.tmp. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = kicad_utils.open_file_write(sys.argv[2], 'wb')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, file=sys.stderr)
    f = sys.stdout
    canOpenFile = False

# Output a set of rows for a header providing general information
html = html.replace('<!--SOURCE-->', net.getSource())
html = html.replace('<!--DATE-->', net.getDate())
html = html.replace('<!--TOOL-->', net.getTool())
html = html.replace('<!--COMPCOUNT-->', "<b>Component Count:</b>" + \
    str(len(net.components)))

row  = "<tr><th style='width:640px'>Ref</th>" + "<th>Qnty</th>"
row += "<th>Value</th>" + "<th>Part</th>" + "<th>Footprint</th>"
row += "<th>Description</th>" + "<th>Vendor</th>" + "<th>DNP</th></tr>"

html = html.replace('<!--TABLEROW-->', row + "<!--TABLEROW-->")

components = net.getInterestingComponents( excludeBOM=True )

# Get all of the components in groups of matching parts + values
# (see kicad_netlist_reader.py)
grouped = net.groupComponents(components)

# Output all of the component information
for group in grouped:
    refs = ""

    # Add the reference of every component in the group and keep a reference
    # to the component so that the other data can be filled in once per group
    for component in group:
        if len(refs) > 0:
            refs += ", "
        refs += component.getRef()
        c = component

    row = "\n    "
    row += "<tr><td>" + refs +"</td><td>" + str(len(group))
    row += "</td><td>" + c.getValue() + "</td><td>" + c.getLibName() + ":"
    row += c.getPartName() + "</td><td>" + c.getFootprint() + "</td><td>"
    row += c.getDescription() + "</td><td>" + c.getField("Vendor") + "</td><td>"
    row += c.getDNPString()
    row += "</td></tr>"

    html = html.replace('<!--TABLEROW-->', row + "<!--TABLEROW-->")

# Write the formatted html to the file
if sys.version_info[0] < 3:
    f.write(html)
else:
    f.write(html.encode('utf-8'))
f.close

if not canOpenFile:
    sys.exit(1)


#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# A basic fp wizard to test arcs in footprints
import pcbnew
import FootprintWizardBase

class DemoArc(FootprintWizardBase.FootprintWizard):
    """Test for arc"""
    GetName = lambda self: "DemoArc"
    GetDescription = lambda self: DemoArc.__doc__
    GetReferencePrefix = lambda self: "demo"
    GetValue = lambda self: "Arc"

    def GenerateParameterList(self):
        self.AddParam("demo", "radius", self.uMM, 8, min_value=0.1)
        self.AddParam("demo", "centerX", self.uMM, 0)
        self.AddParam("demo", "centerY", self.uMM, 0)
        self.AddParam("demo", "angle", self.uDegrees, 90)

    def BuildThisFootprint(self):
        radius = self.parameters["demo"]["radius"]
        arc_angle_deg = self.parameters["demo"]["angle"]
        centerx = self.parameters["demo"]["centerX"]
        centery = self.parameters["demo"]["centerY"]
        startptx = centerx
        startpty = centery+radius
        self.draw.Arc(centerx, centery, startptx, startpty, pcbnew.EDA_ANGLE( arc_angle_deg, pcbnew.DEGREES_T ) )
        t_size = self.GetTextSize()
        self.draw.Reference(0, -t_size, t_size)
        self.draw.Value(0, t_size, t_size)

    def CheckParameters(self):
        pass

DemoArc().register()



#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2024 Jarrett Rainier
# Copyright (C) 2012-2024 KiCad Developers
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, you may find one here:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# or you may search the http://www.gnu.org website for the version 2 license,
# or you may write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
# Based upon the work in:
# Microchip AN2934 Self-Capacitance Sensors
#

import pcbnew
import FootprintWizardBase
import pcbnew
import math, cmath

class ScrollWheelWizard(FootprintWizardBase.FootprintWizard):

    def GetName(self):
        """
        Return footprint name.
        This is specific to each footprint class, you need to implement this
        """
        return 'ScrollWheel'

    def GetDescription(self):
        """
        Return footprint description.
        This is specific to each footprint class, you need to implement this
        """
        return 'Capacitive ScrollWheel wizard'

    def GetValue(self):
        return "ScrollWheel-{od:g}x{id:g}mm".format(
            od = pcbnew.ToMM(self.pads['outer_diameter']),
            id = pcbnew.ToMM(self.pads['inner_diameter'])
            )

    def GenerateParameterList(self):
        self.AddParam("Pads", "steps", self.uInteger, 3, min_value=2)
        self.AddParam("Pads", "bands", self.uInteger, 4, min_value=2)
        self.AddParam("Pads", "outer_diameter", self.uMM, 40)
        self.AddParam("Pads", "inner_diameter", self.uMM, 12)
        self.AddParam("Pads", "deadzone", self.uMM, 4)
        self.AddParam("Pads", "corner_radius", self.uMM, 1)
        self.AddParam("Pads", "clearance", self.uMM, 0.5, min_value=0.5, max_value=1.5)
        self.AddParam("Pads", "edge_silkscreen", self.uBool, True)
        self.AddParam("Pads", "full_silkscreen", self.uBool, True)
    @property
    def pads(self):
        return self.parameters['Pads']
        
    def smdCircle(self,radius,pos):
        arc_angle_deg = 360
        centerx = pos[0]
        centery = pos[1]
        startptx = centerx
        startpty = centery+radius
        self.draw.Arc(centerx, centery, startptx, startpty, pcbnew.EDA_ANGLE( arc_angle_deg, pcbnew.DEGREES_T ) )
        
    def smdArc(self,start,end,pos):
        pos = complex(pos[0], pos[1])
        centerx = pos.real
        centery = pos.imag
        start_angle_rad = cmath.phase(start - pos)
        end_angle_rad = cmath.phase(end - pos)
        start_angle_deg = math.degrees(start_angle_rad)
        end_angle_deg = math.degrees(end_angle_rad)
        arc_angle_deg = (end_angle_deg - start_angle_deg) % 360
        self.draw.Arc(centerx, centery, start.real, start.imag, pcbnew.EDA_ANGLE( arc_angle_deg, pcbnew.DEGREES_T ) )
       
    # Generate points along the arc and create a polygon from those points
    def arc_points(self, start, end, pos, cw=True):
        pos = complex(pos[0], pos[1])
        centerx = pos.real
        centery = pos.imag
        if cw == False:
            start, end = end, start
        start_angle_rad = cmath.phase(start - pos)
        end_angle_rad = cmath.phase(end - pos)
        arc_angle_rad = (end_angle_rad - start_angle_rad) % (2 * math.pi)
        num_points = 20
        
        shape = pcbnew.SHAPE_LINE_CHAIN()
        for i in range(num_points - 1):
            if cw:
                angle = start_angle_rad + (i + 1) * arc_angle_rad / (num_points - 1)
            else:
                angle = start_angle_rad + (num_points - (i + 1)) * arc_angle_rad / (num_points - 1)
            x = centerx + abs(start - pos) * math.cos(angle)
            y = centery + abs(start - pos) * math.sin(angle)
            shape.Append(pcbnew.VECTOR2I(int(x), int(y)))

        return shape
        
    # Find the centre, given two points and a radius
    def find_arc_center(self, p1, p2, radius, cw=True):
        x1, y1 = p1.real, p1.imag
        x2, y2 = p2.real, p2.imag

        mx, my = (x1 + x2) / 2, (y1 + y2) / 2
        d = math.sqrt((x2 - x1)**2 + (y2 - y1)**2)

        if d > 2 * radius:
            raise ValueError("Radius {} is too small to pass through both points {}.".format(radius, d))

        h = math.sqrt(radius**2 - (d / 2)**2)
        dx, dy = (y2 - y1) / d, -(x2 - x1) / d

        if cw:
            cx, cy = mx - h * dx, my - h * dy
        else:
            cx, cy = mx + h * dx, my + h * dy

        return (cx, cy)
        
    # Find the centre, given two points that are at 180 degrees
    def find_avg_centre(self, p1, p2):
        pt1_r, pt1_phi = cmath.polar(p1)
        pt2_r, pt2_phi = cmath.polar(p2)
        
        avg_r = (pt1_r + pt2_r) / 2
        avg_phi = (pt1_phi + pt2_phi) / 2
        
        centre = cmath.rect(avg_r, avg_phi)

        return (centre.real, centre.imag)
        
    def find_avg_radius(self, p1, p2, cw = True):
        pt1_r, pt1_phi = cmath.polar(p1)
        pt2_r, pt2_phi = cmath.polar(p2)
        avg_r = (pt1_r + pt2_r) / 2

        return self.find_arc_center(p1, p2, avg_r, cw)        
        
    def create_pads(self, outer_diameter, inner_diameter, deadzone, corner_radius, steps, bands, clearance):
        outer_radius = outer_diameter / 2
        inner_radius = inner_diameter / 2

        # Angle step for each radial section (each deadzone gap)
        angle_step = 2 * math.pi / steps
        
        # Right side corners, determines tangential OD-ID range
        inner_corner_r = inner_radius + corner_radius
        corner_x = deadzone / 2 + corner_radius
        y_min = cmath.sqrt(inner_corner_r**2 - corner_x**2)
        outer_corner_r = outer_radius - corner_radius
        y_max = cmath.sqrt(outer_corner_r**2 - corner_x**2)
        
        y_range = y_max - y_min
        
        points_right = []
        points_left = []

        for step in range(steps):
            angle = step * angle_step
            points_right.append([])
            points_left.append([])
            for i in range(bands):
                # Positions for right side arc corners (on the top/bot edges of the band)
                corner_y = y_min + (i * y_range / (bands - 1))
                corner_r, corner_phi = cmath.polar(complex(corner_x, corner_y))
                corner = cmath.rect(corner_r, corner_phi + angle)
                right = corner.real, -corner.imag
                
                p1 = cmath.rect(corner_r + corner_radius, corner_phi + angle)
                p1 = p1.real - p1.imag * 1j
                p2 = cmath.rect(corner_r - corner_radius, corner_phi + angle)
                p2 = p2.real - p2.imag * 1j
                points_right[step].append([p1, p2, right])

                # Positions for left side arc corners (centered vertically within the band)
                if i < bands - 1:
                    band_min = y_min + (i * y_range / (bands - 1))
                    band_max = y_min + ((i + 1) * y_range / (bands - 1))
                    corner_y = band_min + (band_max - band_min) / 2
                    corner_r, corner_phi = cmath.polar(complex(-1 * corner_x, corner_y))
                    corner = cmath.rect(corner_r, corner_phi + angle)
                    left = corner.real, -corner.imag
                    
                    p1 = cmath.rect(corner_r - corner_radius, corner_phi + angle)
                    p1 = p1.real - p1.imag * 1j
                    p2 = cmath.rect(corner_r + corner_radius, corner_phi + angle)
                    p2 = p2.real - p2.imag * 1j
                    points_left[step].append([p1, p2, left])
                    

        # Format for first and second args, p1 and p2: 012
        # 0 = "step" radially around the circle
        # 1 = "band" tangentially through the circle
        # 2 = inside or outside point of the arc
        # Third arg is switch/case statement described below
        # Fourth arg cw (T) / ccw (F) for arc drawing
        ordered_pairs = []
        
        # ID edge
        ordered_pairs.append([points_right[1][0][1], points_right[0][0][1], 0, True])
        
        # Moving tangentially, from ID to OD, on right side
        for i in range(bands - 1):
            ordered_pairs.append([points_right[0][i][1], points_right[0][i][0], 1, True])
            ordered_pairs.append([points_right[0][i][0], points_left[steps - 1][i][0], 2, True])
            ordered_pairs.append([points_left[steps - 1][i][0], points_left[steps - 1][i][1], 1, False])
            ordered_pairs.append([points_left[steps - 1][i][1], points_right[0][i + 1][1], 2, False])
            
        # OD edge
        ordered_pairs.append([points_right[0][bands - 1][1], points_right[0][bands - 1][0], 1, True])
        ordered_pairs.append([points_right[0][bands - 1][0], points_right[1][bands - 1][0], 0, False])
        
        for i in range(bands - 1):
            a = bands - (i + 1)
            ordered_pairs.append([points_right[1][a][0], points_right[1][a][1], 1, False])
            ordered_pairs.append([points_right[1][a][1], points_left[0][a - 1][1], 2, True])
            ordered_pairs.append([points_left[0][a - 1][1], points_left[0][a - 1][0], 1, True])
            ordered_pairs.append([points_left[0][a - 1][0], points_right[1][a - 1][0], 2, False])
            ordered_pairs.append([points_right[1][a - 1][0], points_right[1][a - 1][1], 1, False])
  
        shape = pcbnew.SHAPE_LINE_CHAIN()
        for i in range(len(ordered_pairs)):
            step = ordered_pairs[i]
            centre = (step[0].real, step[0].imag)
            
             # Switch case:
             #    centre = 0,
             #    between the two points (ie. corners), or 
             #    perpendicular and average of the radius (ie. "spikes") 
            if ordered_pairs[i][2] == 0:
                centre = (0,0)
            elif ordered_pairs[i][2] == 1:
                centre = self.find_avg_centre(ordered_pairs[i][0], ordered_pairs[i][1])
            elif ordered_pairs[i][2] == 2:
                centre = self.find_avg_radius(ordered_pairs[i][0], ordered_pairs[i][1], ordered_pairs[i][3])
                
            shape.Append(self.arc_points(ordered_pairs[i][0], ordered_pairs[i][1], centre, ordered_pairs[i][3]))
            

        offset = int((outer_radius - inner_radius) / 2 + inner_radius)
        pad = pcbnew.PAD(self.module)
        pad.SetShape(pcbnew.PAD_SHAPE_CUSTOM)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_SMD)
        pad.SetSize(pcbnew.VECTOR2I(deadzone // 2, deadzone // 2))
        pad.SetPosition(pcbnew.VECTOR2I(0, -offset))
        fcuSet = pcbnew.LSET()
        fcuSet.AddLayer(pcbnew.F_Cu)
        pad.SetLayerSet(fcuSet)
        poly = pcbnew.SHAPE_POLY_SET(shape)
        poly.Deflate(int(clearance / 2), pcbnew.CORNER_STRATEGY_ROUND_ALL_CORNERS, int(clearance / 10))
        poly.Move((pcbnew.VECTOR2I(0, offset)))
        pad.AddPrimitive(poly, 0)
        
        for i in range(steps):
            angle_step = (i * 2 * math.pi / steps)
            pos = cmath.rect(offset, angle_step + (math.pi / 2))
            step_pad = pad.ClonePad()
            step_pad.SetName(str(i + 1))
            step_pad.SetOrientation(pcbnew.EDA_ANGLE(angle_step, pcbnew.RADIANS_T ))
            step_pad.SetPosition(pcbnew.VECTOR2I(int(pos.real), int(-pos.imag)))
            self.module.Add(step_pad)
        
    def draw_silkscreen_arcs(self, outer_diameter, inner_diameter, deadzone, corner_radius, steps, bands):
        outer_radius = outer_diameter / 2
        inner_radius = inner_diameter / 2

        # Angle step for each radial section (each deadzone gap)
        angle_step = 2 * math.pi / steps
        
        # Right side
        inner_corner_r = inner_radius + corner_radius
        corner_x = deadzone / 2 + corner_radius
        y_min = cmath.sqrt(inner_corner_r**2 - corner_x**2)
        outer_corner_r = outer_radius - corner_radius
        y_max = cmath.sqrt(outer_corner_r**2 - corner_x**2)
        
        y_range = y_max - y_min
        
        outer_spikes = []
        inner_spikes = []
        right_corners = []
        right_corners_ccw = []
        left_corners = []

        for step in range(steps):
            angle = step * angle_step
            outer_spikes.append([])
            inner_spikes.append([])
            right_corners.append([])
            right_corners_ccw.append([])
            left_corners.append([])
            
            for i in range(bands):
                # Positions for right side arc corners (on the top/bot edges of the band)
                corner_y = y_min + (i * y_range / (bands - 1))
                corner_r, corner_phi = cmath.polar(complex(corner_x, corner_y))
                corner = cmath.rect(corner_r, corner_phi + angle)
                right = corner.real, -corner.imag
                
                outer_spike_corner_r = cmath.rect(corner_r + corner_radius, corner_phi + angle)
                outer_spike_corner_r = outer_spike_corner_r.real - outer_spike_corner_r.imag * 1j
                
                inner_spike_corner_r = cmath.rect(corner_r - corner_radius, corner_phi + angle)
                inner_spike_corner_r = inner_spike_corner_r.real - inner_spike_corner_r.imag * 1j
                
                self.smdArc(inner_spike_corner_r, outer_spike_corner_r,right)
                right_corners[step].append(self.arc_points(inner_spike_corner_r, outer_spike_corner_r,right))
                right_corners_ccw[step].append(self.arc_points(inner_spike_corner_r, outer_spike_corner_r,right, False))

                # Positions for left side arc corners (centered vertically within the band)
                if i < bands:
                    band_min = y_min + (i * y_range / (bands - 1))
                    band_max = y_min + ((i + 1) * y_range / (bands - 1))
                    corner_y = band_min + (band_max - band_min) / 2
                    corner_r, corner_phi = cmath.polar(complex(-1 * corner_x, corner_y))
                    corner = cmath.rect(corner_r, corner_phi + angle)
                    left = corner.real, -corner.imag
                    
                    outer_spike_corner_l = cmath.rect(corner_r - corner_radius, corner_phi + angle)
                    outer_spike_corner_l = outer_spike_corner_l.real - outer_spike_corner_l.imag * 1j
                    outer_spikes[step].append([outer_spike_corner_r, outer_spike_corner_l])
                # if i > 0:
                inner_spike_corner_l = cmath.rect(corner_r + corner_radius, corner_phi + angle)
                inner_spike_corner_l = inner_spike_corner_l.real - inner_spike_corner_l.imag * 1j
                inner_spikes[step].append([inner_spike_corner_r, inner_spike_corner_l])
                if (i + 1) < bands:
                    self.smdArc(inner_spike_corner_l, outer_spike_corner_l,left)
                    left_corners[step].append(self.arc_points(inner_spike_corner_l, outer_spike_corner_l,left, False))

        for step in range(steps):
            id1 = inner_spikes[step][0][0]
            id2 = inner_spikes[(step + 1) % steps][0][0]
            shape = self.arc_points(id2, id1,(0,0))
            for i in range(bands - 1):
                self.smdArc(id1, id2,(0,0))
                if step == 0:
                    shape.Append(right_corners[step][i])
                spike_corner_r = outer_spikes[step][i][0]
                spike_corner_l = outer_spikes[(step - 1) % steps][i][1]
                radius_r = abs(spike_corner_r)
                radius_l = abs(spike_corner_l)
                radius = (radius_r + radius_l) / 2
                spike_centre = self.find_arc_center(spike_corner_r, spike_corner_l, radius, cw=True)
                self.smdArc(spike_corner_r, spike_corner_l,spike_centre)
                if step == 0:
                    shape.Append(self.arc_points(spike_corner_r, spike_corner_l,spike_centre, True))
                    shape.Append(left_corners[(step - 1) % steps][i])
                spike_corner_r = inner_spikes[step][i + 1][0]
                spike_corner_l = inner_spikes[(step - 1) % steps][i][1]
                radius_r = abs(spike_corner_r)
                radius_l = abs(spike_corner_l)
                radius = (radius_r + radius_l) / 2
                spike_centre = self.find_arc_center(spike_corner_r, spike_corner_l, radius, cw=True)
                self.smdArc(spike_corner_r, spike_corner_l,spike_centre)
                if step == 0:
                    shape.Append(self.arc_points(spike_corner_r, spike_corner_l,spike_centre, False))
            shape.Append(right_corners[step][bands - 1])
            od2 = outer_spikes[step][bands - 1][0]
            od1 = outer_spikes[(step + 1) % steps][bands - 1][0]
            
            if step == 0:
                shape.Append(self.arc_points(od1, od2,(0,0), False))
            self.smdArc(od1, od2,(0,0))
                    
    def CheckParameters(self):
        od = pcbnew.ToMM(self.parameters['Pads']['outer_diameter'])
        id = pcbnew.ToMM(self.parameters['Pads']['inner_diameter'])
        
        # Diametral height
        max_height = 20 * 2
        min_height = 8 * 2
        
        self.CheckParam('Pads','outer_diameter',max_value=id+max_height,info="Electrode height must be less than 20mm")
        self.CheckParam('Pads','outer_diameter',min_value=id+min_height,info="Electrode height must be at least 8mm")
        
        self.CheckParam('Pads','inner_diameter',min_value=od-max_height,info="Electrode height must be less than 20mm")
        self.CheckParam('Pads','inner_diameter',max_value=od-min_height,info="Electrode height must be at least 8mm")

    def BuildThisFootprint(self):
        param_steps             = self.pads["steps"]
        param_bands             = self.pads["bands"]
        param_od                = self.pads["outer_diameter"]
        param_id                = self.pads["inner_diameter"]
        param_deadzone          = self.pads["deadzone"]
        param_corner_radius     = self.pads["corner_radius"]
        param_clearance         = self.pads["clearance"]
        param_ss_full           = self.pads["full_silkscreen"]
        param_ss_edge           = self.pads["edge_silkscreen"]

        step_length = float(param_od) / float(param_steps)

        t_size = self.GetTextSize()
        w_text = self.draw.GetLineThickness()
        ypos = param_od/2 + t_size/2 + w_text
        self.draw.Value(0, -ypos, t_size)
        ypos += t_size + w_text*2
        self.draw.Reference(0, -ypos, t_size)

        self.module.SetAttributes(pcbnew.FP_SMD)

        pos = pcbnew.VECTOR2I(0, 0)
        
        if param_ss_edge == True and param_ss_full == False:
            self.smdCircle(param_od / 2,pos)
            self.smdCircle(param_id / 2,pos)
        if param_ss_full == True:
            self.draw_silkscreen_arcs(param_od, param_id, param_deadzone, param_corner_radius, param_steps, param_bands)
        self.create_pads(param_od, param_id, param_deadzone, param_corner_radius, param_steps, param_bands, param_clearance)


ScrollWheelWizard().register()



# -*- coding: utf-8 -*-
"""KiCad Python Shell.

This module provides the python shell for KiCad.

KiCad starts the shell once, by calling makePcbnewShellWindow() the
first time it is opened, subsequently the shell window is just hidden
or shown, as per user requirements.

IF makePcbnewShellWindow() is called again, a second/third shell window
can be created.

Note:
**DO NOT** import pcbnew module if not called from Pcbnew: on msys2 it creates a serious issue:
the python script import is broken because the pcbnew application is not running, and for instance
Pgm() returns a nullptr and Kicad crashes when Pgm is invoked.

"""

import wx
import sys
import os
import pcbnew

from wx.py import crust, version, dispatcher

from .kicad_pyeditor import KiCadEditorNotebookFrame
from .kicad_pyeditor import KiCadEditorNotebook

class KiCadPyShell(KiCadEditorNotebookFrame):
    def __init__(self, parent):
        # Search if a pcbnew frame is open, because import pcbnew can be made only if it exists.
        # frame names are "SchematicFrame" and "PcbFrame"
        #
        # Note we must do this before the KiCadEditorNotebookFrame __init__ call because it ends up calling _setup back on us
        frame = wx.FindWindowByName( "PcbFrame" )

        self.isPcbframe = frame is not None

        KiCadEditorNotebookFrame.__init__(self, parent)

    def _setup_startup(self):
        if self.config_dir != "":
            """Initialise the startup script."""
            # Create filename for startup script.
            self.startup_file = os.path.join(self.config_dir,
                                             "PyShell_pcbnew_startup.py")
            self.execStartupScript = True

            # Check if startup script exists
            if not os.path.isfile(self.startup_file):
                # Not, so try to create a default.
                try:
                    default_startup = open(self.startup_file, 'w')
                    # provide the content for the default startup file.
                    default_startup.write(
                        "### DEFAULT STARTUP FILE FOR KiCad Python Shell\n" +
                        "# Enter any Python code you would like to execute when" +
                        " the PCBNEW python shell first runs.\n" +
                        "\n" +
                        "# For example, uncomment the following lines to import the current board\n" +
                        "\n" +
                        "# import pcbnew\n" +
                        "# import eeschema\n" +
                        "# board = pcbnew.GetBoard()\n" +
                        "# sch = eeschema.GetSchematic()\n")
                    default_startup.close()
                except:
                    pass
        else:
            self.startup_file = ""
            self.execStartupScript = False

    def _setup(self):
        """
        Setup prior to first buffer creation.

        Called automatically by base class during init.
        """
        self.notebook = KiCadEditorNotebook(parent=self.parent)
        intro = 'Py %s' % version.VERSION

        intro += """\n\nDeprecation Notice:\n
This SWIG-based Python interface to the PCB editor is deprecated and will be removed 
in a future version of KiCad.  Please plan to move to the new IPC API and/or make 
use of the kicad-cli tool for your KiCad automation needs.\n\n"""

        import types
        import builtins
        module = types.ModuleType('__main__')
        module.__dict__['__builtins__'] = builtins
        namespace = module.__dict__.copy()

        '''
        Import pcbnew **only** if the board editor exists, to avoid strange behavior if not.
        pcbnew.SETTINGS_MANAGER should be in fact called only if the python console is created
        from the board editor, and if created from schematic editor, should use something like
        eeschema.SETTINGS_MANAGER
        '''
        if self.isPcbframe:
            import pcbnew
            self.config_dir = pcbnew.SETTINGS_MANAGER.GetUserSettingsPath()
        else:
            self.config_dir = ""

        self.dataDir = self.config_dir

        self._setup_startup()
        self.history_file = os.path.join(self.config_dir,
                                         "PyShell_pcbnew.history")

        self.config = None
        # self.config_file = os.path.join(self.config_dir,
        #                                 "PyShell_pcbnew.cfg")
        # self.config = wx.FileConfig(localFilename=self.config_file)
        # self.config.SetRecordDefaults(True)

        self.autoSaveSettings = False
        self.autoSaveHistory = False
        self.LoadSettings()

        self.crust = crust.Crust(parent=self.notebook,
                                 intro=intro, locals=namespace,
                                 rootLabel="locals()",
                                 startupScript=self.startup_file,
                                 execStartupScript=self.execStartupScript)

        self.crust._CheckShouldSplit()
        self.shell = self.crust.shell
        # Override the filling so that status messages go to the status bar.
        self.crust.filling.tree.setStatusText = self.parent.SetStatusText
        # Override the shell so that status messages go to the status bar.
        self.shell.setStatusText = self.parent.SetStatusText
        # Fix a problem with the sash shrinking to nothing.
        self.crust.filling.SetSashPosition(200)
        self.notebook.AddPage(page=self.crust, text='*Shell*', select=True)
        self.setEditor(self.crust.editor)
        self.crust.editor.SetFocus()

        self.LoadHistory()

    def OnAbout(self, event):
        """Display an About window."""
        title = 'About : KiCad - Python Shell'
        text = "Enhanced Python Shell for KiCad\n\n" + \
               "KiCad Revision: %s\n" % pcbnew.FullVersion() + \
               "Platform: %s\n" % sys.platform + \
               "Python Version: %s\n" % sys.version.split()[0] + \
               "wxPython Version: %s\n" % wx.VERSION_STRING + \
               ("\t(%s)\n" % ", ".join(wx.PlatformInfo[1:]))

        dialog = wx.MessageDialog(self.parent, text, title,
                                  wx.OK | wx.ICON_INFORMATION)
        dialog.ShowModal()
        dialog.Destroy()

    def EditStartupScript(self):
        """Open a Edit buffer of the startup script file."""
        self.bufferCreate(filename=self.startup_file)

    def LoadSettings(self):
        """Load settings for the shell."""
        if self.config is not None:
            KiCadEditorNotebookFrame.LoadSettings(self.parent, self.config)
            self.autoSaveSettings = \
                self.config.ReadBool('Options/AutoSaveSettings', False)
            self.execStartupScript = \
                self.config.ReadBool('Options/ExecStartupScript', True)
            self.autoSaveHistory = \
                self.config.ReadBool('Options/AutoSaveHistory', False)
            self.hideFoldingMargin = \
                self.config.ReadBool('Options/HideFoldingMargin', True)

    def SaveSettings(self, force=False):
        """
        Save settings for the shell.

        Arguments:

        force -- False - Autosaving.  True - Manual Saving.
        """
        if self.config is not None:
            # always save these
            self.config.WriteBool('Options/AutoSaveSettings',
                                  self.autoSaveSettings)
            if self.autoSaveSettings or force:
                KiCadEditorNotebookFrame.SaveSettings(self, self.config)

                self.config.WriteBool('Options/AutoSaveHistory',
                                      self.autoSaveHistory)
                self.config.WriteBool('Options/ExecStartupScript',
                                      self.execStartupScript)
                self.config.WriteBool('Options/HideFoldingMargin',
                                      self.hideFoldingMargin)
            if self.autoSaveHistory:
                self.SaveHistory()

    def DoSaveSettings(self):
        """Menu function to trigger saving the shells settings."""
        if self.config is not None:
            self.SaveSettings(force=True)
            self.config.Flush()

    def SaveHistory(self):
        """Save shell history to the shell history file."""
        if self.dataDir:
            try:
                name = self.history_file
                f = file(name, 'w')
                hist = []
                enc = wx.GetDefaultPyEncoding()
                for h in self.shell.history:
                    if isinstance(h, unicode):
                        h = h.encode(enc)
                    hist.append(h)
                hist = '\x00\n'.join(hist)
                f.write(hist)
                f.close()
            except:
                d = wx.MessageDialog(self, "Error saving history file.",
                                     "Error", wx.ICON_EXCLAMATION | wx.OK)
                d.ShowModal()
                d.Destroy()
                raise

    def LoadHistory(self):
        """Load shell history from the shell history file."""
        if self.dataDir:
            name = self.history_file
            if os.path.exists(name):
                try:
                    f = file(name, 'U')
                    hist = f.read()
                    f.close()
                    self.shell.history = hist.split('\x00\n')
                    dispatcher.send(signal="Shell.loadHistory",
                                    history=self.shell.history)
                except:
                    d = wx.MessageDialog(self,
                                         "Error loading history file!",
                                         "Error", wx.ICON_EXCLAMATION | wx.OK)
                    d.ShowModal()
                    d.Destroy()


def makePcbnewShellWindow(parentid):
    """
    Create a new Shell Window and return its handle.

    Arguments:
    parent -- The parent window to attach to.

    Returns:
    The handle to the new window.
    """

    parent = wx.FindWindowById( parentid )

    frmname = parent.GetName()
    return KiCadPyShell(parent)



#  Copyright 2019-2020 Julian Fellinger
#  Copyright (C) 2023 KiCad Developers.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import FootprintWizardBase
import pcbnew

class MutualcapButtonWizard(FootprintWizardBase.FootprintWizard):

    def GetName(self):
        """
        Return footprint name.
        This is specific to each footprint class, you need to implement this
        """
        return "Mutualcap Touch Button"

    def GetDescription(self):
        """
        Return footprint description.
        This is specific to each footprint class, you need to implement this
        """
        return "Wizard for Mutual Capacitance Touch Buttons."

    def GetValue(self):
        return "TouchButton-{width:g}x{height:g}mm".format(
            width = pcbnew.ToMM(self.pads["Width"]),
            height= pcbnew.ToMM(self.pads["Height"])
            )

    def GenerateParameterList(self):
        self.AddParam("Pads", "Width", self.uMM, 12)
        self.AddParam("Pads", "Height", self.uMM, 12)
        self.AddParam("Pads", "Outer electrode width", self.uMM, 1)
        self.AddParam("Pads", "Inner electrode width", self.uMM, 0.1)
        self.AddParam("Pads", "Draw line around button", self.uBool, True)

    @property
    def pads(self):
        return self.parameters["Pads"]


    # build a rectangular pad
    def smdRectPad(self,module,size,pos,name):
        pad = pcbnew.PAD(module)
        pad.SetSize(size)
        pad.SetShape(pcbnew.PAD_SHAPE_RECT)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_SMD)

        layerset = pcbnew.LSET()
        layerset.AddLayer(pcbnew.F_Cu)
        pad.SetLayerSet(layerset)

        pad.SetPosition(pos)
        pad.SetName(name)
        return pad

    # This method checks the parameters provided to wizard and set errors
    def CheckParameters(self):
        pass

    def BuildThisFootprint(self):
        # o refers to the outer (drive) electrode
        # i refers to the inner (receive) electrode
        w   = self.pads["Width"]
        h   = self.pads["Height"]
        oew = self.pads["Outer electrode width"]
        iew = self.pads["Inner electrode width"]
        drawBox = self.pads["Draw line around button"]

        oFingerCount = int((w-3*oew-iew)/(1.5*oew+iew))
        oBorderWidth = (w-oew-iew-oFingerCount*(1.5*oew+iew))/2
        clearance    = oew/2

        ###border h
        self.module.SetLayer(pcbnew.F_Cu)
        size_pad = pcbnew.VECTOR2I( w, oew )
        self.module.Add(self.smdRectPad(self.module,size_pad,
                                        pcbnew.VECTOR2I(0, (int)((-h/2)+oew/2) ),
                                        str(1)))
        self.module.Add(self.smdRectPad(self.module,size_pad,
                                        pcbnew.VECTOR2I(0, (int)(( h/2)-oew/2) ),
                                        str(1)))
        ###border v
        size_pad = pcbnew.VECTOR2I( int(oBorderWidth), int(h-2*oew) )
        self.module.Add(self.smdRectPad(self.module,size_pad,
                                        pcbnew.VECTOR2I( (int)((-w/2)+oBorderWidth/2), 0),
                                        str(1)))
        self.module.Add(self.smdRectPad(self.module,size_pad,
                                        pcbnew.VECTOR2I( (int)((w/2)-oBorderWidth/2), 0),
                                        str(1)))

        xPos = -w/2+oBorderWidth+clearance+iew
        oFingerSize = pcbnew.VECTOR2I( int(oew/2), int(h-oew-iew-2*clearance-oew) )
        iFingerSize = pcbnew.VECTOR2I( int(iew), int(h-2*oew-2*clearance) )
        #horizontal inner electrode trace
        self.module.Add(self.smdRectPad(self.module,pcbnew.VECTOR2I( int(w-2*oBorderWidth-2*clearance), int(iew) ),
                        pcbnew.VECTOR2I(0, (int)(h/2-oew-clearance-iew/2) ),str(2)))
        for i in range(0,oFingerCount):
            #inner electrode fingers
            self.module.Add(self.smdRectPad(self.module,iFingerSize,
                                            pcbnew.VECTOR2I( (int)(xPos-iew/2) ,0),
                                            str(2)))
            #outer electrode fingers
            self.module.Add(self.smdRectPad(self.module,oFingerSize,
                                            pcbnew.VECTOR2I( (int)(xPos+oew*0.75),
                                                             (int)(-w/h-oew-iew/2+oew/2) ),
                                            str(1)))
            xPos += oew*1.5+iew

        #rightmost inner electrode finger
        self.module.Add(self.smdRectPad(self.module,iFingerSize,
                                        pcbnew.VECTOR2I( (int)(xPos-iew/2), 0),str(2)))

        textSize = self.GetTextSize()
        self.draw.Value(0, h/2+textSize, textSize)
        self.draw.Reference(0, -h/2-textSize, textSize)

        # Add a extra text (${REFERENCE}) on the F_Fab layer
        extra_text = pcbnew.PCB_TEXT( self.module )
        extra_text.SetLayer( pcbnew.F_Fab )
        extra_text.SetPosition( pcbnew.VECTOR2I( 0, 0) )
        extra_text.SetTextSize( pcbnew.VECTOR2I( textSize, textSize ) )
        extra_text.SetText( "${REFERENCE}" )
        self.module.Add( extra_text )

        #optionally draw silkscreen line around button
        if(drawBox):
            self.draw.SetLayer( pcbnew.F_SilkS )
            self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) )     #Default per KLC F5.3
            self.draw.Box( 0, 0, w, h )

        # draw coutyard  around button
        self.draw.SetLayer( pcbnew.F_CrtYd )
        self.draw.SetLineThickness( pcbnew.FromMM( 0.05 ) )     #Default per KLC F5.3
        margin = pcbnew.FromMM( 0.25 )  #Default per KLC F5.3
        self.draw.Box( 0, 0, w + margin*2 , h + margin*2 )

MutualcapButtonWizard().register()



#
# Example python script to generate a BOM from a KiCad generic netlist
#
# Example: Sorted and Grouped CSV BOM
#

"""
    @package
    Output: CSV (comma-separated)
    Grouped By: Value, Symbol Name, Footprint, DNP
    Sorted By: Ref
    Fields: Item, Qty, Reference(s), Value, LibPart, Footprint, Datasheet, DNP, all additional symbol fields

    Outputs ungrouped components first, then outputs grouped components.

    Command line:
    python "pathToFile/bom_csv_grouped_by_value.py" "%I" "%O.csv"
"""

from __future__ import print_function

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import kicad_utils
import csv
import sys

# A helper function to filter/convert a string read in netlist
#currently: do nothing
def fromNetlistText( aText ):
    return aText

def myEqu(self, other):
    """myEqu is a more advanced equivalence function for components which is
    used by component grouping. Normal operation is to group components based
    on their value and footprint.

    In this example of a custom equivalency operator we compare the
    value, the part name and the footprint.
    """
    result = True
    if self.getValue() != other.getValue():
        result = False
    elif self.getPartName() != other.getPartName():
        result = False
    elif self.getFootprint() != other.getFootprint():
        result = False
    elif self.getDNP() != other.getDNP():
        result = False

    return result

# Override the component equivalence operator - it is important to do this
# before loading the netlist, otherwise all components will have the original
# equivalency operator.
kicad_netlist_reader.comp.__eq__ = myEqu

if len(sys.argv) != 3:
    print("Usage ", __file__, "<generic_netlist.xml> <output.csv>", file=sys.stderr)
    sys.exit(1)


# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = kicad_utils.open_file_writeUTF8(sys.argv[2], 'w')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print( __file__, ":", e, sys.stderr )
    f = sys.stdout
    canOpenFile = False

# subset the components to those wanted in the BOM, controlled
# by <configure> block in kicad_netlist_reader.py
components = net.getInterestingComponents( excludeBOM=True )

compfields = net.gatherComponentFieldUnion(components)
partfields = net.gatherLibPartFieldUnion()

# remove Reference, Value, Datasheet, and Footprint, they will come from 'columns' below
partfields -= set( ['Reference', 'Value', 'Datasheet', 'Footprint'] )

columnset = compfields | partfields     # union

# prepend an initial 'hard coded' list and put the enchillada into list 'columns'
hardcoded_columns = ['Item', 'Qty', 'Reference(s)', 'Value', 'LibPart', 'Footprint', 'Datasheet', 'DNP']
columns = hardcoded_columns + sorted(list(columnset))

# Create a new csv writer object to use as the output formatter
out = csv.writer( f, lineterminator='\n', delimiter=',', quotechar='\"', quoting=csv.QUOTE_ALL )

# override csv.writer's writerow() to support encoding conversion (initial encoding is utf8):
def writerow( acsvwriter, columns ):
    utf8row = []
    for col in columns:
        utf8row.append( fromNetlistText( str(col) ) )
    acsvwriter.writerow( utf8row )

# Output a set of rows as a header providing general information
writerow( out, ['Source:', net.getSource()] )
writerow( out, ['Date:', net.getDate()] )
writerow( out, ['Tool:', net.getTool()] )
writerow( out, ['Generator:', sys.argv[0]] )
writerow( out, ['Component Count:', len(components)] )
writerow( out, [] )

writerow( out, ['Collated Components:'] )
writerow( out, [] )                        # blank line
writerow( out, columns )                   # reuse same columns

row=[]

# Get all of the components in groups of matching parts + values
# (see kicad_netlist_reader.py)
grouped = net.groupComponents(components)


# Output component information organized by group, aka as collated:
item = 0
for group in grouped:
    del row[:]
    refs = ""
    refs_l = []

    # Add the reference of every component in the group and keep a reference
    # to the component so that the other data can be filled in once per group
    for component in group:
        refs_l.append( component.getRef() )
        c = component
    
    refs = ", ".join(refs_l)
    
    # Fill in the component groups common data
    # columns = ['Item', 'Qty', 'Reference(s)', 'Value', 'LibPart', 'Footprint', 'Datasheet', 'DNP'] + sorted(list(columnset))
    item += 1
    row.append( item )
    row.append( len(group) )
    row.append( refs );
    row.append( c.getValue() )
    row.append( c.getLibName() + ":" + c.getPartName() )
    row.append( net.getGroupFootprint(group) )
    row.append( net.getGroupDatasheet(group) )
    row.append( c.getDNPString() )

    # add user fields by name
    for field in columns[len(hardcoded_columns):]:
        row.append( net.getGroupField(group, field) );

    writerow( out, row  )

f.close()

if not canOpenFile:
    sys.exit(1)



'''
Provides the backend for a basic python editor in KiCad.

This takes most code from PyShell/PyCrust but adapts it to the KiCad
environment where the Python doesn't create a frame but instead hooks
into the existing KIWAY_PLAYER

Original PyCrust code used from
https://github.com/wxWidgets/Phoenix/tree/master/wx/py
'''

import wx

from wx.py import crust, version, dispatcher, editwindow
from wx.py.editor import Editor, openSingle, openMultiple, saveSingle, messageDialog
from wx.py.buffer import Buffer
from wx import stc

ID_NEW = wx.ID_NEW
ID_OPEN = wx.ID_OPEN
ID_REVERT = wx.ID_REVERT
ID_CLOSE = wx.ID_CLOSE
ID_SAVE = wx.ID_SAVE
ID_SAVEAS = wx.ID_SAVEAS
ID_PRINT = wx.ID_PRINT
ID_EXIT = wx.ID_EXIT
ID_UNDO = wx.ID_UNDO
ID_REDO = wx.ID_REDO
ID_CUT = wx.ID_CUT
ID_COPY = wx.ID_COPY
ID_PASTE = wx.ID_PASTE
ID_CLEAR = wx.ID_CLEAR
ID_SELECTALL = wx.ID_SELECTALL
ID_EMPTYBUFFER = 1
ID_ABOUT = wx.ID_ABOUT
ID_HELP = 1
ID_AUTOCOMP_SHOW = 1
ID_AUTOCOMP_MAGIC = 1
ID_AUTOCOMP_SINGLE = 1
ID_AUTOCOMP_DOUBLE = 1
ID_CALLTIPS_SHOW = 1
ID_CALLTIPS_INSERT = 1
ID_COPY_PLUS = 1
ID_NAMESPACE = 1
ID_PASTE_PLUS = 1
ID_WRAP = 1
ID_TOGGLE_MAXIMIZE = 1
ID_SHOW_LINENUMBERS = 1
ID_ENABLESHELLMODE = 1
ID_ENABLEAUTOSYMPY = 1
ID_AUTO_SAVESETTINGS = 1
ID_SAVEACOPY = 1
ID_SAVEHISTORY = 1
ID_SAVEHISTORYNOW = 1
ID_CLEARHISTORY = 1
ID_SAVESETTINGS = 1
ID_DELSETTINGSFILE = 1
ID_EDITSTARTUPSCRIPT = 1
ID_EXECSTARTUPSCRIPT = 1
ID_SHOWPYSLICESTUTORIAL = 1
ID_FIND = wx.ID_FIND
ID_FINDNEXT = 1
ID_FINDPREVIOUS = 1
ID_SHOWTOOLS = 1
ID_HIDEFOLDINGMARGIN = 1

INTRO = "KiCad - Python Shell"

class KiCadPyFrame():

    def __init__(self, parent):
        """Create a Frame instance."""

        self.parent = parent
        self.parent.CreateStatusBar()
        self.parent.SetStatusText('Frame')
        self.shellName='KiCad Python Editor'
        self.__createMenus()

        self.iconized = False
        self.findDlg = None
        self.findData = wx.FindReplaceData()
        self.findData.SetFlags(wx.FR_DOWN)

        self.parent.Bind(wx.EVT_ICONIZE, self.OnIconize)

    def SetIDs( self ):
        ID_EMPTYBUFFER = self.parent.NewControlId()
        ID_HELP = self.parent.NewControlId()
        ID_AUTOCOMP_SHOW = self.parent.NewControlId()
        ID_AUTOCOMP_MAGIC = self.parent.NewControlId()
        ID_AUTOCOMP_SINGLE = self.parent.NewControlId()
        ID_AUTOCOMP_DOUBLE = self.parent.NewControlId()
        ID_CALLTIPS_SHOW = self.parent.NewControlId()
        ID_CALLTIPS_INSERT = self.parent.NewControlId()
        ID_COPY_PLUS = self.parent.NewControlId()
        ID_NAMESPACE = self.parent.NewControlId()
        ID_PASTE_PLUS = self.parent.NewControlId()
        ID_WRAP = self.parent.NewControlId()
        ID_TOGGLE_MAXIMIZE = self.parent.NewControlId()
        ID_SHOW_LINENUMBERS = self.parent.NewControlId()
        ID_ENABLESHELLMODE = self.parent.NewControlId()
        ID_ENABLEAUTOSYMPY = self.parent.NewControlId()
        ID_AUTO_SAVESETTINGS = self.parent.NewControlId()
        ID_SAVEACOPY = self.parent.NewControlId()
        ID_SAVEHISTORY = self.parent.NewControlId()
        ID_SAVEHISTORYNOW = self.parent.NewControlId()
        ID_CLEARHISTORY = self.parent.NewControlId()
        ID_SAVESETTINGS = self.parent.NewControlId()
        ID_DELSETTINGSFILE = self.parent.NewControlId()
        ID_EDITSTARTUPSCRIPT = self.parent.NewControlId()
        ID_EXECSTARTUPSCRIPT = self.parent.NewControlId()
        ID_SHOWPYSLICESTUTORIAL = self.parent.NewControlId()
        ID_FINDNEXT = self.parent.NewControlId()
        ID_FINDPREVIOUS = self.parent.NewControlId()
        ID_SHOWTOOLS = self.parent.NewControlId()
        ID_HIDEFOLDINGMARGIN = self.parent.NewControlId()

    def OnIconize(self, event):
        """Event handler for Iconize."""
        self.iconized = event.IsIconized()


    def __createMenus(self):
        # File Menu
        m = self.fileMenu = wx.Menu()
        m.Append(ID_NEW, '&New \tCtrl+N',
                 'New file')
        m.Append(ID_OPEN, '&Open... \tCtrl+O',
                 'Open file')
        m.AppendSeparator()
        m.Append(ID_REVERT, '&Revert \tCtrl+R',
                 'Revert to last saved version')
        m.Append(ID_CLOSE, '&Close \tCtrl+W',
                 'Close file')
        m.AppendSeparator()
        m.Append(ID_SAVE, '&Save... \tCtrl+S',
                 'Save file')
        m.Append(ID_SAVEAS, 'Save &As \tCtrl+Shift+S',
                 'Save file with new name')
        m.AppendSeparator()
        m.Append(ID_PRINT, '&Print... \tCtrl+P',
                 'Print file')
        m.AppendSeparator()
        m.Append(ID_NAMESPACE, '&Update Namespace \tCtrl+Shift+N',
                 'Update namespace for autocompletion and calltips')
        m.AppendSeparator()
        m.Append(ID_EXIT, 'E&xit\tCtrl+Q', 'Exit Program')

        # Edit
        m = self.editMenu = wx.Menu()
        m.Append(ID_UNDO, '&Undo \tCtrl+Z',
                 'Undo the last action')
        m.Append(ID_REDO, '&Redo \tCtrl+Y',
                 'Redo the last undone action')
        m.AppendSeparator()
        m.Append(ID_CUT, 'Cu&t \tCtrl+X',
                 'Cut the selection')
        m.Append(ID_COPY, '&Copy \tCtrl+C',
                 'Copy the selection')
        m.Append(ID_COPY_PLUS, 'Cop&y Plus \tCtrl+Shift+C',
                 'Copy the selection - retaining prompts')
        m.Append(ID_PASTE, '&Paste \tCtrl+V', 'Paste from clipboard')
        m.Append(ID_PASTE_PLUS, 'Past&e Plus \tCtrl+Shift+V',
                 'Paste and run commands')
        m.AppendSeparator()
        m.Append(ID_CLEAR, 'Cle&ar',
                 'Delete the selection')
        m.Append(ID_SELECTALL, 'Select A&ll \tCtrl+A',
                 'Select all text')
        m.AppendSeparator()
        m.Append(ID_EMPTYBUFFER, 'E&mpty Buffer...',
                 'Delete all the contents of the edit buffer')
        m.Append(ID_FIND, '&Find Text... \tCtrl+F',
                 'Search for text in the edit buffer')
        m.Append(ID_FINDNEXT, 'Find &Next \tCtrl+G',
                 'Find next instance of the search text')
        m.Append(ID_FINDPREVIOUS, 'Find Pre&vious \tCtrl+Shift+G',
                 'Find previous instance of the search text')

        # View
        m = self.viewMenu = wx.Menu()
        m.Append(ID_WRAP, '&Wrap Lines\tCtrl+Shift+W',
                 'Wrap lines at right edge', wx.ITEM_CHECK)
        m.Append(ID_SHOW_LINENUMBERS, '&Show Line Numbers\tCtrl+Shift+L',
                 'Show Line Numbers', wx.ITEM_CHECK)
        m.Append(ID_TOGGLE_MAXIMIZE, '&Toggle Maximize\tF11',
                 'Maximize/Restore Application')

        # # Options
        # m = self.optionsMenu = wx.Menu()
        #
        # self.historyMenu = wx.Menu()
        # self.historyMenu.Append(ID_SAVEHISTORY, '&Autosave History',
        #          'Automatically save history on close', wx.ITEM_CHECK)
        # self.historyMenu.Append(ID_SAVEHISTORYNOW, '&Save History Now',
        #          'Save history')
        # self.historyMenu.Append(ID_CLEARHISTORY, '&Clear History ',
        #          'Clear history')
        # m.AppendSubMenu(self.historyMenu, "&History", "History Options")
        #
        # self.startupMenu = wx.Menu()
        # self.startupMenu.Append(ID_EXECSTARTUPSCRIPT,
        #                         'E&xecute Startup Script',
        #                         'Execute Startup Script', wx.ITEM_CHECK)
        # self.startupMenu.Append(ID_EDITSTARTUPSCRIPT,
        #                         '&Edit Startup Script...',
        #                         'Edit Startup Script')
        # m.AppendSubMenu(self.startupMenu, '&Startup', 'Startup Options')

        self.settingsMenu = wx.Menu()
        self.settingsMenu.Append(ID_AUTO_SAVESETTINGS,
                                 '&Auto Save Settings',
                                 'Automatically save settings on close', wx.ITEM_CHECK)
        self.settingsMenu.Append(ID_SAVESETTINGS,
                                 '&Save Settings',
                                 'Save settings now')
        self.settingsMenu.Append(ID_DELSETTINGSFILE,
                                 '&Revert to default',
                                 'Revert to the default settings')
        m.AppendSubMenu(self.settingsMenu, '&Settings', 'Settings Options')

        m = self.helpMenu = wx.Menu()
        m.Append(ID_HELP, '&Help\tF1', 'Help!')
        m.AppendSeparator()
        m.Append(ID_ABOUT, '&About...', 'About this program')

        b = self.menuBar = wx.MenuBar()
        b.Append(self.fileMenu, '&File')
        b.Append(self.editMenu, '&Edit')
        b.Append(self.viewMenu, '&View')
        # b.Append(self.optionsMenu, '&Options')
        b.Append(self.helpMenu, '&Help')
        self.parent.SetMenuBar(b)

        self.parent.Bind(wx.EVT_MENU, self.OnFileNew, id=ID_NEW)
        self.parent.Bind(wx.EVT_MENU, self.OnFileOpen, id=ID_OPEN)
        self.parent.Bind(wx.EVT_MENU, self.OnFileRevert, id=ID_REVERT)
        self.parent.Bind(wx.EVT_MENU, self.OnFileClose, id=ID_CLOSE)
        self.parent.Bind(wx.EVT_MENU, self.OnFileSave, id=ID_SAVE)
        self.parent.Bind(wx.EVT_MENU, self.OnFileSaveAs, id=ID_SAVEAS)
        self.parent.Bind(wx.EVT_MENU, self.OnFileSaveACopy, id=ID_SAVEACOPY)
        self.parent.Bind(wx.EVT_MENU, self.OnFileUpdateNamespace, id=ID_NAMESPACE)
        self.parent.Bind(wx.EVT_MENU, self.OnFilePrint, id=ID_PRINT)
        self.parent.Bind(wx.EVT_MENU, self.OnExit, id=ID_EXIT)
        self.parent.Bind(wx.EVT_MENU, self.OnUndo, id=ID_UNDO)
        self.parent.Bind(wx.EVT_MENU, self.OnRedo, id=ID_REDO)
        self.parent.Bind(wx.EVT_MENU, self.OnCut, id=ID_CUT)
        self.parent.Bind(wx.EVT_MENU, self.OnCopy, id=ID_COPY)
        self.parent.Bind(wx.EVT_MENU, self.OnCopyPlus, id=ID_COPY_PLUS)
        self.parent.Bind(wx.EVT_MENU, self.OnPaste, id=ID_PASTE)
        self.parent.Bind(wx.EVT_MENU, self.OnPastePlus, id=ID_PASTE_PLUS)
        self.parent.Bind(wx.EVT_MENU, self.OnClear, id=ID_CLEAR)
        self.parent.Bind(wx.EVT_MENU, self.OnSelectAll, id=ID_SELECTALL)
        self.parent.Bind(wx.EVT_MENU, self.OnEmptyBuffer, id=ID_EMPTYBUFFER)
        self.parent.Bind(wx.EVT_MENU, self.OnAbout, id=ID_ABOUT)
        self.parent.Bind(wx.EVT_MENU, self.OnHelp, id=ID_HELP)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoCompleteShow, id=ID_AUTOCOMP_SHOW)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoCompleteMagic, id=ID_AUTOCOMP_MAGIC)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoCompleteSingle, id=ID_AUTOCOMP_SINGLE)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoCompleteDouble, id=ID_AUTOCOMP_DOUBLE)
        self.parent.Bind(wx.EVT_MENU, self.OnCallTipsShow, id=ID_CALLTIPS_SHOW)
        self.parent.Bind(wx.EVT_MENU, self.OnCallTipsInsert, id=ID_CALLTIPS_INSERT)
        self.parent.Bind(wx.EVT_MENU, self.OnWrap, id=ID_WRAP)
        self.parent.Bind(wx.EVT_MENU, self.OnToggleMaximize, id=ID_TOGGLE_MAXIMIZE)
        self.parent.Bind(wx.EVT_MENU, self.OnShowLineNumbers, id=ID_SHOW_LINENUMBERS)
        self.parent.Bind(wx.EVT_MENU, self.OnEnableShellMode, id=ID_ENABLESHELLMODE)
        self.parent.Bind(wx.EVT_MENU, self.OnEnableAutoSympy, id=ID_ENABLEAUTOSYMPY)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoSaveSettings, id=ID_AUTO_SAVESETTINGS)
        self.parent.Bind(wx.EVT_MENU, self.OnSaveHistory, id=ID_SAVEHISTORY)
        self.parent.Bind(wx.EVT_MENU, self.OnSaveHistoryNow, id=ID_SAVEHISTORYNOW)
        self.parent.Bind(wx.EVT_MENU, self.OnClearHistory, id=ID_CLEARHISTORY)
        self.parent.Bind(wx.EVT_MENU, self.OnSaveSettings, id=ID_SAVESETTINGS)
        self.parent.Bind(wx.EVT_MENU, self.OnDelSettingsFile, id=ID_DELSETTINGSFILE)
        self.parent.Bind(wx.EVT_MENU, self.OnEditStartupScript, id=ID_EDITSTARTUPSCRIPT)
        self.parent.Bind(wx.EVT_MENU, self.OnExecStartupScript, id=ID_EXECSTARTUPSCRIPT)
        self.parent.Bind(wx.EVT_MENU, self.OnShowPySlicesTutorial, id=ID_SHOWPYSLICESTUTORIAL)
        self.parent.Bind(wx.EVT_MENU, self.OnFindText, id=ID_FIND)
        self.parent.Bind(wx.EVT_MENU, self.OnFindNext, id=ID_FINDNEXT)
        self.parent.Bind(wx.EVT_MENU, self.OnFindPrevious, id=ID_FINDPREVIOUS)
        self.parent.Bind(wx.EVT_MENU, self.OnToggleTools, id=ID_SHOWTOOLS)
        self.parent.Bind(wx.EVT_MENU, self.OnHideFoldingMargin, id=ID_HIDEFOLDINGMARGIN)

        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_NEW)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_OPEN)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_REVERT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_CLOSE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SAVE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SAVEAS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_NAMESPACE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_PRINT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_UNDO)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_REDO)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_CUT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_COPY)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_COPY_PLUS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_PASTE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_PASTE_PLUS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_CLEAR)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SELECTALL)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_EMPTYBUFFER)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_AUTOCOMP_SHOW)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_AUTOCOMP_MAGIC)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_AUTOCOMP_SINGLE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_AUTOCOMP_DOUBLE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_CALLTIPS_SHOW)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_CALLTIPS_INSERT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_WRAP)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SHOW_LINENUMBERS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_ENABLESHELLMODE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_ENABLEAUTOSYMPY)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_AUTO_SAVESETTINGS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SAVESETTINGS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_DELSETTINGSFILE)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_EXECSTARTUPSCRIPT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SHOWPYSLICESTUTORIAL)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SAVEHISTORY)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SAVEHISTORYNOW)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_CLEARHISTORY)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_EDITSTARTUPSCRIPT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_FIND)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_FINDNEXT)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_FINDPREVIOUS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_SHOWTOOLS)
        self.parent.Bind(wx.EVT_UPDATE_UI, self.OnUpdateMenu, id=ID_HIDEFOLDINGMARGIN)

        self.parent.Bind(wx.EVT_ACTIVATE, self.OnActivate)
        self.parent.Bind(wx.EVT_FIND, self.OnFindNext)
        self.parent.Bind(wx.EVT_FIND_NEXT, self.OnFindNext)
        self.parent.Bind(wx.EVT_FIND_CLOSE, self.OnFindClose)

    def OnShowLineNumbers(self, event):
        win = wx.Window.FindFocus()
        if hasattr(win, 'lineNumbers'):
            win.lineNumbers = event.IsChecked()
            win.setDisplayLineNumbers(win.lineNumbers)

    def OnFileNew(self, event):
        self.bufferNew()

    def OnFileOpen(self, event):
        self.bufferOpen()

    def OnFileRevert(self, event):
        self.bufferRevert()

    def OnFileClose(self, event):
        self.bufferClose()

    def OnFileSave(self, event):
        self.bufferSave()

    def OnFileSaveAs(self, event):
        self.bufferSaveAs()

    def OnFileSaveACopy(self, event):
        self.bufferSaveACopy()

    def OnFileUpdateNamespace(self, event):
        self.updateNamespace()

    def OnFilePrint(self, event):
        self.bufferPrint()

    def OnExit(self, event):
        self.parent.Close(False)

    def OnUndo(self, event):
        win = wx.Window.FindFocus()
        win.Undo()

    def OnRedo(self, event):
        win = wx.Window.FindFocus()
        win.Redo()

    def OnCut(self, event):
        win = wx.Window.FindFocus()
        win.Cut()

    def OnCopy(self, event):
        win = wx.Window.FindFocus()
        win.Copy()

    def OnCopyPlus(self, event):
        win = wx.Window.FindFocus()
        win.CopyWithPrompts()

    def OnPaste(self, event):
        win = wx.Window.FindFocus()
        win.Paste()

    def OnPastePlus(self, event):
        win = wx.Window.FindFocus()
        win.PasteAndRun()

    def OnClear(self, event):
        win = wx.Window.FindFocus()
        win.Clear()

    def OnEmptyBuffer(self, event):
        win = wx.Window.FindFocus()
        d = wx.MessageDialog(self,
                             "Are you sure you want to clear the edit buffer,\n"
                             "deleting all the text?",
                             "Empty Buffer", wx.OK | wx.CANCEL | wx.ICON_QUESTION)
        answer = d.ShowModal()
        d.Destroy()
        if (answer == wx.ID_OK):
            win.ClearAll()
            if hasattr(win,'prompt'):
                win.prompt()

    def OnSelectAll(self, event):
        win = wx.Window.FindFocus()
        win.SelectAll()

    def OnAbout(self, event):
        """Display an About window."""
        title = 'About'
        text = 'Your message here.'
        dialog = wx.MessageDialog(self.parent, text, title,
                                  wx.OK | wx.ICON_INFORMATION)
        dialog.ShowModal()
        dialog.Destroy()

    def OnHelp(self, event):
        """Display a Help window."""
        title = 'Help'
        text = "Type 'shell.help()' in the shell window."
        dialog = wx.MessageDialog(self.parent, text, title,
                                  wx.OK | wx.ICON_INFORMATION)
        dialog.ShowModal()
        dialog.Destroy()

    def OnAutoCompleteShow(self, event):
        win = wx.Window.FindFocus()
        win.autoComplete = event.IsChecked()

    def OnAutoCompleteMagic(self, event):
        win = wx.Window.FindFocus()
        win.autoCompleteIncludeMagic = event.IsChecked()

    def OnAutoCompleteSingle(self, event):
        win = wx.Window.FindFocus()
        win.autoCompleteIncludeSingle = event.IsChecked()

    def OnAutoCompleteDouble(self, event):
        win = wx.Window.FindFocus()
        win.autoCompleteIncludeDouble = event.IsChecked()

    def OnCallTipsShow(self, event):
        win = wx.Window.FindFocus()
        win.autoCallTip = event.IsChecked()

    def OnCallTipsInsert(self, event):
        win = wx.Window.FindFocus()
        win.callTipInsert = event.IsChecked()

    def OnWrap(self, event):
        win = wx.Window.FindFocus()
        win.SetWrapMode(event.IsChecked())
        wx.CallLater(1, self.shell.EnsureCaretVisible)

    def OnToggleMaximize(self, event):
        self.parent.Maximize(not self.parent.IsMaximized())

    def OnSaveHistory(self, event):
        self.autoSaveHistory = event.IsChecked()

    def OnSaveHistoryNow(self, event):
        self.SaveHistory()

    def OnClearHistory(self, event):
        self.shell.clearHistory()

    def OnEnableShellMode(self, event):
        self.enableShellMode = event.IsChecked()

    def OnEnableAutoSympy(self, event):
        self.enableAutoSympy = event.IsChecked()

    def OnHideFoldingMargin(self, event):
        self.hideFoldingMargin = event.IsChecked()

    def OnAutoSaveSettings(self, event):
        self.autoSaveSettings = event.IsChecked()

    def OnSaveSettings(self, event):
        self.DoSaveSettings()

    def OnDelSettingsFile(self, event):
        if self.config is not None:
            d = wx.MessageDialog(
                self.parent, "Do you want to revert to the default settings?\n" +
                "A restart is needed for the change to take effect",
                "Warning", wx.OK | wx.CANCEL | wx.ICON_QUESTION)
            answer = d.ShowModal()
            d.Destroy()
            if (answer == wx.ID_OK):
                self.config.DeleteAll()
                self.LoadSettings()


    def OnEditStartupScript(self, event):
        if hasattr(self, 'EditStartupScript'):
            self.EditStartupScript()

    def OnExecStartupScript(self, event):
        self.execStartupScript = event.IsChecked()
        self.SaveSettings(force=True)

    def OnShowPySlicesTutorial(self,event):
        self.showPySlicesTutorial = event.IsChecked()
        self.SaveSettings(force=True)

    def OnFindText(self, event):
        if self.findDlg is not None:
            return
        win = wx.Window.FindFocus()
        if self.shellName == 'PyCrust':
            self.findDlg = wx.FindReplaceDialog(win, self.findData,
                                               "Find",wx.FR_NOWHOLEWORD)
        else:
            self.findDlg = wx.FindReplaceDialog(win, self.findData,
                "Find & Replace", wx.FR_NOWHOLEWORD|wx.FR_REPLACEDIALOG)
        self.findDlg.Show()

    def OnFindNext(self, event,backward=False):
        if backward and (self.findData.GetFlags() & wx.FR_DOWN):
            self.findData.SetFlags( self.findData.GetFlags() ^ wx.FR_DOWN )
        elif not backward and not (self.findData.GetFlags() & wx.FR_DOWN):
            self.findData.SetFlags( self.findData.GetFlags() ^ wx.FR_DOWN )

        if not self.findData.GetFindString():
            self.OnFindText(event)
            return
        if isinstance(event, wx.FindDialogEvent):
            win = self.findDlg.GetParent()
        else:
            win = wx.Window.FindFocus()
        win.DoFindNext(self.findData, self.findDlg)
        if self.findDlg is not None:
            self.OnFindClose(None)

    def OnFindPrevious(self, event):
        self.OnFindNext(event,backward=True)

    def OnFindClose(self, event):
        self.findDlg.Destroy()
        self.findDlg = None

    def OnToggleTools(self, event):
        self.ToggleTools()


    def OnUpdateMenu(self, event):
        """Update menu items based on current status and context."""
        win = wx.Window.FindFocus()
        id = event.GetId()
        event.Enable(True)
        try:
            if id == ID_NEW:
                event.Enable(hasattr(self, 'bufferNew'))
            elif id == ID_OPEN:
                event.Enable(hasattr(self, 'bufferOpen'))
            elif id == ID_REVERT:
                event.Enable(hasattr(self, 'bufferRevert')
                             and self.hasBuffer())
            elif id == ID_CLOSE:
                event.Enable(hasattr(self, 'bufferClose')
                             and self.hasBuffer())
            elif id == ID_SAVE:
                event.Enable(hasattr(self, 'bufferSave')
                             and self.bufferHasChanged())
            elif id == ID_SAVEAS:
                event.Enable(hasattr(self, 'bufferSaveAs')
                             and self.hasBuffer())
            elif id == ID_SAVEACOPY:
                event.Enable(hasattr(self, 'bufferSaveACopy')
                             and self.hasBuffer())
            elif id == ID_NAMESPACE:
                event.Enable(hasattr(self, 'updateNamespace')
                             and self.hasBuffer())
            elif id == ID_PRINT:
                event.Enable(hasattr(self, 'bufferPrint')
                             and self.hasBuffer())
            elif id == ID_UNDO:
                event.Enable(win.CanUndo())
            elif id == ID_REDO:
                event.Enable(win.CanRedo())
            elif id == ID_CUT:
                event.Enable(win.CanCut())
            elif id == ID_COPY:
                event.Enable(win.CanCopy())
            elif id == ID_COPY_PLUS:
                event.Enable(win.CanCopy() and hasattr(win, 'CopyWithPrompts'))
            elif id == ID_PASTE:
                event.Enable(win.CanPaste())
            elif id == ID_PASTE_PLUS:
                event.Enable(win.CanPaste() and hasattr(win, 'PasteAndRun'))
            elif id == ID_CLEAR:
                event.Enable(win.CanCut())
            elif id == ID_SELECTALL:
                event.Enable(hasattr(win, 'SelectAll'))
            elif id == ID_EMPTYBUFFER:
                event.Enable(hasattr(win, 'ClearAll') and not win.GetReadOnly())
            elif id == ID_AUTOCOMP_SHOW:
                event.Check(win.autoComplete)
            elif id == ID_AUTOCOMP_MAGIC:
                event.Check(win.autoCompleteIncludeMagic)
            elif id == ID_AUTOCOMP_SINGLE:
                event.Check(win.autoCompleteIncludeSingle)
            elif id == ID_AUTOCOMP_DOUBLE:
                event.Check(win.autoCompleteIncludeDouble)
            elif id == ID_CALLTIPS_SHOW:
                event.Check(win.autoCallTip)
            elif id == ID_CALLTIPS_INSERT:
                event.Check(win.callTipInsert)
            elif id == ID_WRAP:
                event.Check(win.GetWrapMode())

            elif id == ID_SHOW_LINENUMBERS:
                event.Check(win.lineNumbers)
            elif id == ID_ENABLESHELLMODE:
                event.Check(self.enableShellMode)
                event.Enable(self.config is not None)
            elif id == ID_ENABLEAUTOSYMPY:
                event.Check(self.enableAutoSympy)
                event.Enable(self.config is not None)
            # elif id == ID_AUTO_SAVESETTINGS:
            #     event.Check(self.autoSaveSettings)
            #     event.Enable(self.config is not None)
            # elif id == ID_SAVESETTINGS:
            #     event.Enable(self.config is not None and
            #                  hasattr(self, 'DoSaveSettings'))
            # elif id == ID_DELSETTINGSFILE:
            #     event.Enable(self.config is not None)
            #
            # elif id == ID_EXECSTARTUPSCRIPT:
            #     event.Check(self.execStartupScript)
            #     event.Enable(self.config is not None)
            #
            # elif id == ID_SAVEHISTORY:
            #     event.Check(self.autoSaveHistory)
            #     event.Enable(self.dataDir is not None)
            # elif id == ID_SAVEHISTORYNOW:
            #     event.Enable(self.dataDir is not None and
            #                  hasattr(self, 'SaveHistory'))
            # elif id == ID_CLEARHISTORY:
            #     event.Enable(self.dataDir is not None)
            #
            # elif id == ID_EDITSTARTUPSCRIPT:
            #     event.Enable(hasattr(self, 'EditStartupScript'))
            #     event.Enable(self.dataDir is not None)

            elif id == ID_FIND:
                event.Enable(hasattr(win, 'DoFindNext'))
            elif id == ID_FINDNEXT:
                event.Enable(hasattr(win, 'DoFindNext'))
            elif id == ID_FINDPREVIOUS:
                event.Enable(hasattr(win, 'DoFindNext'))

            elif id == ID_SHOWTOOLS:
                event.Check(self.ToolsShown())

            elif id == ID_HIDEFOLDINGMARGIN:
                event.Check(self.hideFoldingMargin)
                event.Enable(self.config is not None)

            else:
                event.Enable(False)
        except AttributeError:
            # This menu option is not supported in the current context.
            event.Enable(False)


    def OnActivate(self, event):
        """
        Event Handler for losing the focus of the Frame. Should close
        Autocomplete listbox, if shown.
        """
        if not event.GetActive():
            # If autocomplete active, cancel it.  Otherwise, the
            # autocomplete list will stay visible on top of the
            # z-order after switching to another application
            win = wx.Window.FindFocus()
            if hasattr(win, 'AutoCompActive') and win.AutoCompActive():
                win.AutoCompCancel()
        event.Skip()



    def LoadSettings(self, config):
        """Called by derived classes to load settings specific to the Frame"""
        pos  = wx.Point(config.ReadInt('Window/PosX', -1),
                        config.ReadInt('Window/PosY', -1))

        size = wx.Size(config.ReadInt('Window/Width', -1),
                       config.ReadInt('Window/Height', -1))

        self.SetSize(size)
        self.Move(pos)


    def SaveSettings(self, config):
        """Called by derived classes to save Frame settings to a wx.Config object"""

        # TODO: track position/size so we can save it even if the
        # frame is maximized or iconized.
        if not self.iconized and not self.parent.IsMaximized():
            w, h = self.GetSize()
            config.WriteInt('Window/Width', w)
            config.WriteInt('Window/Height', h)

            px, py = self.GetPosition()
            config.WriteInt('Window/PosX', px)
            config.WriteInt('Window/PosY', py)

class KiCadEditorFrame(KiCadPyFrame):
    def __init__(self, parent=None, id=-1, title='KiCad Python'):

        """Create EditorFrame instance."""
        KiCadPyFrame.__init__(self, parent)
        self.buffers = {}
        self.buffer = None  # Current buffer.
        self.editor = None
        self._defaultText = title
        self._statusText = self._defaultText
        self.parent.SetStatusText(self._statusText)
        self.parent.Bind( wx.EVT_IDLE, self.OnIdle )
        self.parent.Bind( wx.EVT_CLOSE, self.OnClose )
        self._setup()

    def _setup(self):
        """Setup prior to first buffer creation.

        Useful for subclasses."""
        pass

    def setEditor(self, editor):
        self.editor = editor
        self.buffer = self.editor.buffer
        self.buffers[self.buffer.id] = self.buffer

    def OnClose(self, event):
        """Event handler for closing."""
        for buffer in self.buffers.values():
            self.buffer = buffer
            if buffer.hasChanged():
                cancel = self.bufferSuggestSave()
                if cancel and event.CanVeto():
                    event.Veto()
                    return

        self.parent.Hide()
        pass

    def OnIdle(self, event):
        """Event handler for idle time."""
        self._updateStatus()
        if hasattr(self, 'notebook'):
            self._updateTabText()
        self._updateTitle()
        event.Skip()

    def _updateStatus(self):
        """Show current status information."""
        if self.editor and hasattr(self.editor, 'getStatus'):
            status = self.editor.getStatus()
            text = 'File: %s  |  Line: %d  |  Column: %d' % status
        else:
            text = self._defaultText
        if text != self._statusText:
            self.parent.SetStatusText(text)
            self._statusText = text

    def _updateTabText(self):
        """Show current buffer information on notebook tab."""
##         suffix = ' **'
##         notebook = self.notebook
##         selection = notebook.GetSelection()
##         if selection == -1:
##             return
##         text = notebook.GetPageText(selection)
##         window = notebook.GetPage(selection)
##         if window.editor and window.editor.buffer.hasChanged():
##             if text.endswith(suffix):
##                 pass
##             else:
##                 notebook.SetPageText(selection, text + suffix)
##         else:
##             if text.endswith(suffix):
##                 notebook.SetPageText(selection, text[:len(suffix)])

    def _updateTitle(self):
        """Show current title information."""
        title = self.GetTitle()
        if self.bufferHasChanged():
            if title.startswith('* '):
                pass
            else:
                self.SetTitle('* ' + title)
        else:
            if title.startswith('* '):
                self.SetTitle(title[2:])

    def hasBuffer(self):
        """Return True if there is a current buffer."""
        if self.buffer:
            return True
        else:
            return False

    def bufferClose(self):
        """Close buffer."""
        if self.bufferHasChanged():
            cancel = self.bufferSuggestSave()
            if cancel:
                return cancel
        self.bufferDestroy()
        cancel = False
        return cancel

    def bufferCreate(self, filename=None):
        """Create new buffer."""
        self.bufferDestroy()
        buffer = Buffer()
        self.panel = panel = wx.Panel(parent=self, id=-1)
        panel.Bind (wx.EVT_ERASE_BACKGROUND, lambda x: x)
        editor = Editor(parent=panel)
        panel.editor = editor
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(editor.window, 1, wx.EXPAND)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)
        sizer.Layout()
        buffer.addEditor(editor)
        buffer.open(filename)
        self.setEditor(editor)
        self.editor.setFocus()
        self.SendSizeEvent()


    def bufferDestroy(self):
        """Destroy the current buffer."""
        if self.buffer:
            for editor in self.buffer.editors.values():
                editor.destroy()
            self.editor = None
            del self.buffers[self.buffer.id]
            self.buffer = None
            self.panel.Destroy()


    def bufferHasChanged(self):
        """Return True if buffer has changed since last save."""
        if self.buffer:
            return self.buffer.hasChanged()
        else:
            return False

    def bufferNew(self):
        """Create new buffer."""
        if self.bufferHasChanged():
            cancel = self.bufferSuggestSave()
            if cancel:
                return cancel
        self.bufferCreate()
        cancel = False
        return cancel

    def bufferOpen(self):
        """Open file in buffer."""
        if self.bufferHasChanged():
            cancel = self.bufferSuggestSave()
            if cancel:
                return cancel
        filedir = ''
        if self.buffer and self.buffer.doc.filedir:
            filedir = self.buffer.doc.filedir
        result = openSingle(directory=filedir)
        if result.path:
            self.bufferCreate(result.path)
        cancel = False
        return cancel

    def bufferSave(self):
        """Save buffer to its file."""
        if self.buffer.doc.filepath:
            self.buffer.save()
            cancel = False
        else:
            cancel = self.bufferSaveAs()
        return cancel

    def bufferSaveAs(self):
        """Save buffer to a new filename."""
        if self.bufferHasChanged() and self.buffer.doc.filepath:
            cancel = self.bufferSuggestSave()
            if cancel:
                return cancel
        filedir = ''
        if self.buffer and self.buffer.doc.filedir:
            filedir = self.buffer.doc.filedir
        result = saveSingle(directory=filedir)
        if result.path:
            self.buffer.saveAs(result.path)
            cancel = False
        else:
            cancel = True
        return cancel

    def bufferSuggestSave(self):
        """Suggest saving changes.  Return True if user selected Cancel."""
        result = messageDialog(parent=None,
                               message='%s has changed.\n'
                                       'Would you like to save it first'
                                       '?' % self.buffer.name,
                               title='Save current file?')
        if result.positive:
            cancel = self.bufferSave()
        else:
            cancel = result.text == 'Cancel'
        return cancel

    def updateNamespace(self):
        """Update the buffer namespace for autocompletion and calltips."""
        if self.buffer.updateNamespace():
            self.parent.SetStatusText('Namespace updated')
        else:
            self.parent.SetStatusText('Error executing, unable to update namespace')


class KiCadEditorNotebookFrame(KiCadEditorFrame):
    def __init__(self, parent):
        """Create EditorNotebookFrame instance."""

        self.notebook = None
        KiCadEditorFrame.__init__(self, parent)

        if self.notebook:
            """Keep pydoc output on stdout instead of pager and
                place the stdout into the editor window """
            import pydoc, sys
            self._keep_stdin = sys.stdin

            """getline will crash unexpectedly, so we don't support it
                and bold fonts wreak havoc on our output, so strip them as well"""
            pydoc.getpager = lambda: pydoc.plainpager
            pydoc.Helper.getline = lambda self, prompt: None
            pydoc.TextDoc.use_bold = lambda self, text: text

            dispatcher.connect(receiver=self._editorChange,
                               signal='EditorChange', sender=self.notebook)

    def _setup(self):
        """Setup prior to first buffer creation.

        Called automatically by base class during init."""
        self.notebook = EditorNotebook(parent=self)
        intro = 'Py %s' % version.VERSION
        import imp
        module = imp.new_module('__main__')
        module.__dict__['__builtins__'] = __builtins__
        namespace = module.__dict__.copy()
        self.crust = crust.Crust(parent=self.notebook, intro=intro, locals=namespace)
        self.shell = self.crust.shell
        # Override the filling so that status messages go to the status bar.
        self.crust.filling.tree.setStatusText = self.SetStatusText
        # Override the shell so that status messages go to the status bar.
        self.shell.setStatusText = self.SetStatusText
        # Fix a problem with the sash shrinking to nothing.
        self.crust.filling.SetSashPosition(200)
        self.notebook.AddPage(page=self.crust, text='*Shell*', select=True)
        self.setEditor(self.crust.editor)
        self.crust.editor.SetFocus()


    def _editorChange(self, editor):
        """Editor change signal receiver."""
        if not self:
            dispatcher.disconnect(receiver=self._editorChange,
                                  signal='EditorChange', sender=self.notebook)
            return
        self.setEditor(editor)

    def _updateTitle(self):
        """Show current title information."""
        pass


    def bufferCreate(self, filename=None):
        """Create new buffer."""
        buffer = Buffer()
        panel = wx.Panel(parent=self.notebook, id=-1)
        panel.Bind(wx.EVT_ERASE_BACKGROUND, lambda x: x)
        editor = Editor(parent=panel)
        panel.editor = editor
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(editor.window, 1, wx.EXPAND)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)
        sizer.Layout()
        buffer.addEditor(editor)
        buffer.open(filename)
        self.setEditor(editor)
        self.notebook.AddPage(page=panel, text=self.buffer.name, select=True)
        self.editor.setFocus()

    def bufferDestroy(self):
        """Destroy the current buffer."""
        selection = self.notebook.GetSelection()
##         print("Destroy Selection:", selection)
        if selection > 0:  # Don't destroy the PyCrust tab.
            if self.buffer:
                del self.buffers[self.buffer.id]
                self.buffer = None  # Do this before DeletePage().
            self.notebook.DeletePage(selection)

    def bufferNew(self):
        """Create new buffer."""
        self.bufferCreate()
        cancel = False
        return cancel

    def bufferOpen(self):
        """Open file in buffer."""
        filedir = ''
        if self.buffer and self.buffer.doc.filedir:
            filedir = self.buffer.doc.filedir
        result = openMultiple(directory=filedir)
        for path in result.paths:
            self.bufferCreate(path)
        cancel = False
        return cancel


class KiCadEditorNotebook(wx.Notebook):
    """A notebook containing a page for each editor."""

    def __init__(self, parent):
        """Create EditorNotebook instance."""
        wx.Notebook.__init__(self, parent, id=-1, style=wx.CLIP_CHILDREN)
        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGING, self.OnPageChanging, id=self.GetId())
        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.OnPageChanged, id=self.GetId())
        self.Bind(wx.EVT_IDLE, self.OnIdle)

    def OnIdle(self, event):
        """Event handler for idle time."""
        self._updateTabText()
        event.Skip()

    def _updateTabText(self):
        """Show current buffer display name on all but first tab."""
        size = 3
        changed = ' **'
        unchanged = ' --'
        selection = self.GetSelection()
        if selection < 1:
            return
        text = self.GetPageText(selection)
        window = self.GetPage(selection)
        if not window.editor:
            return
        if text.endswith(changed) or text.endswith(unchanged):
            name = text[:-size]
        else:
            name = text
        if name != window.editor.buffer.name:
            text = window.editor.buffer.name
        if window.editor.buffer.hasChanged():
            if text.endswith(changed):
                text = None
            elif text.endswith(unchanged):
                text = text[:-size] + changed
            else:
                text += changed
        else:
            if text.endswith(changed):
                text = text[:-size] + unchanged
            elif text.endswith(unchanged):
                text = None
            else:
                text += unchanged
        if text is not None:
            self.SetPageText(selection, text)
            self.Refresh()  # Needed on Win98.

    def OnPageChanging(self, event):
        """Page changing event handler."""
        event.Skip()

    def OnPageChanged(self, event):
        """Page changed event handler."""
        new = event.GetSelection()
        window = self.GetPage(new)
        dispatcher.send(signal='EditorChange', sender=self,
                        editor=window.editor)
        window.SetFocus()
        event.Skip()


def is_using_dark_background():
    """Backport of wxSystemAppearance::IsUsingDarkBackground() from wxWidgets 3.1."""
    bg = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
    fg = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)

    return get_luminance(fg) - get_luminance(bg) > 0.2


def get_luminance(color):
    """Backport of wxColour::GetLuminance() from wxWidgets 3.1."""
    return (0.299*color.Red() + 0.587*color.Green() + 0.114*color.Blue()) / 255.0


editwindow_old_init = editwindow.EditWindow.__init__

calltip_dark_bg_color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND).ChangeLightness(108)
calltip_dark_fg_color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT).ChangeLightness(108)

def editwindow_new_init(self, *args, **kwargs):
    editwindow_old_init(self, *args, **kwargs)

    # Cannot be used until we switch to wxWidgets >=3.1.3.
    #if wx.SystemSettings.GetAppearance().IsUsingDarkBackground():

    # If a dark theme is detected, we set the default system colors and add our own Python
    # styles (which are not otherwise supplied by the system).
    if is_using_dark_background():
        faces = editwindow.FACES

        # Override the background in all styles by setting the default style, then copying this
        # to all styles with StyleClearAll(). Iterating over all styles would be slower.
        self.StyleSetForeground(stc.STC_STYLE_DEFAULT,
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT))
        self.StyleSetBackground(stc.STC_STYLE_DEFAULT,
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND))
        self.StyleClearAll()

        self.SetSelForeground(True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT))
        self.SetSelBackground(True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT))

        # Override the colors set in EditWindow.setStyles(). We go in the same order as in that
        # function. Most of the foreground colors are just inverses of the default. The background
        # colors default system background color with slightly adjusted brightness.

        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER,
                "back:#3F3F3F,face:%(mono)s,size:%(lnsize)d".format(faces))
        self.StyleSetSpec(stc.STC_STYLE_CONTROLCHAR, "face:%(mono)s".format(faces))

        self.StyleSetSpec(stc.STC_STYLE_BRACELIGHT, "fore:#FFFF00")
        self.StyleSetBackground(stc.STC_STYLE_BRACELIGHT,
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND).ChangeLightness(120))

        self.StyleSetSpec(stc.STC_STYLE_BRACEBAD, "fore:#FF0000")
        self.StyleSetBackground(stc.STC_STYLE_BRACEBAD,
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND).ChangeLightness(120))

        # Python styles
        self.StyleSetSpec(stc.STC_P_DEFAULT, "face:%(mono)s".format(faces))
        self.StyleSetSpec(stc.STC_P_COMMENTLINE,
                "fore:#FF80FF,face:%(mono)s".format(faces))
        self.StyleSetSpec(stc.STC_P_NUMBER, "")
        self.StyleSetSpec(stc.STC_P_STRING, "fore:#80FF80,face:%(mono)s".format(faces))
        self.StyleSetSpec(stc.STC_P_CHARACTER,
                "fore:#80FF80,face:%(mono)s".format(faces))
        self.StyleSetSpec(stc.STC_P_WORD, "fore:#FFFF80,bold")
        self.StyleSetSpec(stc.STC_P_TRIPLE, "fore:#80FFFF")

        self.StyleSetSpec(stc.STC_P_TRIPLEDOUBLE, "fore:#FFFFCC")
        self.StyleSetBackground(stc.STC_P_TRIPLEDOUBLE,
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND).ChangeLightness(105))

        self.StyleSetSpec(stc.STC_P_CLASSNAME, "fore:#FFFF00,bold")
        self.StyleSetSpec(stc.STC_P_DEFNAME, "fore:#FF8080,bold")
        self.StyleSetSpec(stc.STC_P_OPERATOR, "")
        self.StyleSetSpec(stc.STC_P_IDENTIFIER, "")
        self.StyleSetSpec(stc.STC_P_COMMENTBLOCK, "fore:#808080")
        self.StyleSetSpec(stc.STC_P_STRINGEOL,
                "fore:#FFFFFF,face:%(mono)s,back:#1F3F1F,eolfilled".format(faces))

        # Yellow calltip background is too contrasting.
        self.CallTipSetBackground(calltip_dark_bg_color)
        self.CallTipSetForeground(calltip_dark_fg_color)

        # Caret is black by default, override this.
        self.SetCaretForeground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT))


# HACK: Monkey-patch EditWindow to support dark themes.
editwindow.EditWindow.__init__ = editwindow_new_init


calltip_old_init = crust.Calltip.__init__

def calltip_new_init(self, *args, **kwargs):
    calltip_old_init(self, *args, **kwargs)

    if is_using_dark_background():
        self.SetBackgroundColour(calltip_dark_bg_color)
        self.SetForegroundColour(calltip_dark_fg_color)

# HACK: Monkey-patch Calltip to support dark themes.
crust.Calltip.__init__ = calltip_new_init



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from __future__ import division
import pcbnew

import FootprintWizardBase
import PadArray as PA

class QFPWizard(FootprintWizardBase.FootprintWizard):

    def GetName(self):
        return "QFP"

    def GetDescription(self):
        return "Quad Flat Package (QFP) footprint wizard"

    def GenerateParameterList(self):
        self.AddParam("Pads", "n", self.uInteger, 100, multiple=4, min_value=4)
        self.AddParam("Pads", "pitch", self.uMM, 0.5, designator='e')
        self.AddParam("Pads", "width", self.uMM, 0.25, designator='X1')
        self.AddParam("Pads", "length", self.uMM, 1.5, designator='Y1')
        self.AddParam("Pads", "horizontal spacing", self.uMM, 15, designator='C1')
        self.AddParam("Pads", "vertical spacing", self.uMM, 15, designator='C2')
        self.AddParam("Pads", "oval", self.uBool, True)

        self.AddParam("Package", "width", self.uMM, 14, designator='D1')
        self.AddParam("Package", "height", self.uMM, 14, designator='E1')
        self.AddParam("Package", "courtyard margin", self.uMM, 0.25, min_value=0.2)

    @property
    def pads(self):
        return self.parameters['Pads']

    @property
    def package(self):
        return self.parameters['Package']

    def CheckParameters(self):
        # todo - custom checking
        pass

    def GetValue(self):
        return "QFP-{n}_{x:g}x{y:g}_Pitch{p:g}mm".format(
            n = self.pads['n'],
            x = pcbnew.ToMM(self.package['width']),
            y = pcbnew.ToMM(self.package['height']),
            p = pcbnew.ToMM(self.pads['pitch'])
            )

    def BuildThisFootprint(self):

        pad_pitch = self.pads["pitch"]
        pad_length = self.pads["length"]
        pad_width = self.pads["width"]

        v_pitch = self.pads["vertical spacing"]
        h_pitch = self.pads["horizontal spacing"]

        pads_per_row = int(self.pads["n"] // 4)

        row_len = (pads_per_row - 1) * pad_pitch

        pad_shape = pcbnew.PAD_SHAPE_OVAL if self.pads["oval"] else pcbnew.PAD_SHAPE_RECT

        h_pad = PA.PadMaker(self.module).SMDPad( pad_length, pad_width,
                                                 shape=pad_shape, rot_degree=90.0)
        v_pad = PA.PadMaker(self.module).SMDPad( pad_length, pad_width, shape=pad_shape)

        #left row
        pin1Pos = pcbnew.VECTOR2I( (int)(-h_pitch / 2), 0)
        array = PA.PadLineArray(h_pad, pads_per_row, pad_pitch, True, pin1Pos)
        array.SetFirstPadInArray(1)
        array.AddPadsToModule(self.draw)

        #bottom row
        pin1Pos = pcbnew.VECTOR2I( 0, (int)(v_pitch / 2) )
        array = PA.PadLineArray(v_pad, pads_per_row, pad_pitch, False, pin1Pos)
        array.SetFirstPadInArray(pads_per_row + 1)
        array.AddPadsToModule(self.draw)

        #right row
        pin1Pos = pcbnew.VECTOR2I( (int)(h_pitch / 2), 0)
        array = PA.PadLineArray(h_pad, pads_per_row, -pad_pitch, True,
                                pin1Pos)
        array.SetFirstPadInArray(2*pads_per_row + 1)
        array.AddPadsToModule(self.draw)

        #top row
        pin1Pos = pcbnew.VECTOR2I(0, (int)(-v_pitch / 2) )
        array = PA.PadLineArray(v_pad, pads_per_row, -pad_pitch, False,
                                pin1Pos)
        array.SetFirstPadInArray(3*pads_per_row + 1)
        array.AddPadsToModule(self.draw)

        offset = pcbnew.FromMM(0.15)

        x = self.parameters["Package"]["width"] / 2 + offset
        y = self.parameters["Package"]["height"] / 2 + offset
        inner = (row_len / 2) + pad_pitch

        # Add outline to F_Fab layer
        self.draw.SetLayer(pcbnew.F_Fab)
        thick = self.draw.GetLineThickness()

        bevel = min( pcbnew.FromMM(1.0), self.package['width']/2, self.package['height']/2 )

        w = self.package['width']
        h = self.package['height']

        # outermost limits of pins
        right_edge = (h_pitch + pad_length) / 2
        left_edge = -right_edge

        bottom_edge = (v_pitch + pad_length) / 2
        top_edge = -bottom_edge

        self.draw.SetLineThickness( pcbnew.FromMM( 0.1 ) ) #Default per KLC F5.2 as of 12/2018
        self.draw.BoxWithDiagonalAtCorner(0, 0, w, h, bevel)

        # Draw silkscreen
        self.draw.SetLayer( pcbnew.F_SilkS )
        self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) ) #Default per KLC F5.1 as of 12/2018

        #top left - as per IPC-7351C
        self.draw.Polyline([(-inner, -y), (-x, -y), (-x, -inner), (left_edge, -inner)])
        # top right
        self.draw.Polyline([(inner, -y), (x, -y), (x, -inner)])
        # bottom left
        self.draw.Polyline([(-inner, y), (-x, y), (-x, inner)])
        # bottom right
        self.draw.Polyline([(inner, y), (x, y), (x, inner)])

        # Courtyard
        cmargin = self.parameters["Package"]["courtyard margin"]
        self.draw.SetLayer( pcbnew.F_CrtYd )
        sizex = ( right_edge + cmargin ) * 2
        sizey = ( bottom_edge + cmargin ) * 2
        # round size to nearest 0.1mm, rectangle will thus land on a 0.05mm grid
        sizex = pcbnew.PutOnGridMM( sizex, 0.1 )
        sizey = pcbnew.PutOnGridMM( sizey, 0.1 )

        self.draw.SetLineThickness( pcbnew.FromMM( 0.05 ) ) #Default per KLC F5.3 as of 12/2018
        self.draw.Box( 0, 0, sizex, sizey )
        # restore line thickness to previous value
        self.draw.SetLineThickness(pcbnew.FromMM(thick))

        #reference and value
        text_size = self.GetTextSize()  # IPC nominal
        text_offset = v_pitch / 2 + text_size + pad_length / 2

        self.draw.Value(0, text_offset, text_size)
        self.draw.Reference(0, -text_offset, text_size)

        # Add a extra text (${REFERENCE}) on the F_Fab layer
        extra_text = pcbnew.PCB_TEXT( self.module )
        extra_text.SetLayer( pcbnew.F_Fab )
        extra_text.SetPosition( pcbnew.VECTOR2I( 0, 0) )
        extra_text.SetTextSize( pcbnew.VECTOR2I( text_size, text_size ) )
        extra_text.SetText( "${REFERENCE}" )
        self.module.Add( extra_text )

        # set SMD attribute
        self.module.SetAttributes(pcbnew.FP_SMD)

QFPWizard().register()



#
# Example python script to generate a BOM from a KiCad generic netlist
#

"""
    @package
    Output: text file (tab-separated)
    Grouped By: ungrouped, one component per line
    Sorted By: Ref
    Fields: Ref, Value, Part, Footprint, Description, Vendor, DNP

    Command line:
    python "pathToFile/bom_txt_sorted_by_ref.py" "%I" "%O.txt"
"""

from __future__ import print_function

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import kicad_utils
import csv
import sys


# A helper function to filter/convert a string read in netlist
#currently: do nothing
def fromNetlistText( aText ):
    return aText

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = kicad_utils.open_file_writeUTF8(sys.argv[2], 'w' )
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

# Create a new csv writer object to use as the output formatter, although we
# are created a tab delimited list instead!
out = csv.writer(f, lineterminator='\n', delimiter='\t', quotechar="\"", quoting=csv.QUOTE_ALL)

# override csv.writer's writerow() to support utf8 encoding:
def writerow( acsvwriter, columns ):
    utf8row = []
    for col in columns:
        txt=str(col);
        utf8row.append( fromNetlistText(txt) )
    acsvwriter.writerow( utf8row )

components = net.getInterestingComponents( excludeBOM=True )

# Output a field delimited header line
writerow( out, ['Source:', net.getSource()] )
writerow( out, ['Date:', net.getDate()] )
writerow( out, ['Tool:', net.getTool()] )
writerow( out, ['Component Count:', len(components)] )
writerow( out, ['Ref', 'Value', 'Part', 'Footprint', 'Description', 'Vendor', 'DNP'] )

# Output all of the component information
for c in components:
    writerow( out, [c.getRef(), c.getValue(), c.getLibName() + ":" + c.getPartName(),
        c.getFootprint(), c.getDescription(), c.getField("Vendor"), c.getDNPString()])

if not canOpenFile:
    sys.exit(1)


#
# KiCad python module for interpreting generic netlists which can be used
# to generate Bills of materials, etc.
#
# Remember these files use UTF8 encoding
#
# No string formatting is used on purpose as the only string formatting that
# is current compatible with python 2.4+ to 3.0+ is the '%' method, and that
# is due to be deprecated in 3.0+ soon
#

"""
    @package
    Helper module for interpreting generic netlist and build custom
    bom generators or netlists in foreign format
"""


from __future__ import print_function
import sys
import xml.sax as sax
import re
import pdb
import string

#-----<Configure>----------------------------------------------------------------

# excluded_fields is a list of regular expressions.  If any one matches a field
# from either a component or a libpart, then that will not be included as a
# column in the BOM.  Otherwise all columns from all used libparts and components
# will be unionized and will appear.  Some fields are impossible to blacklist, such
# as Ref, Value, Footprint, and Datasheet.  Additionally Qty and Item are supplied
# unconditionally as columns, and may not be removed.
excluded_fields = [
    #'Price@1000'
    ]


# You may exclude components from the BOM by either:
#
# 1) adding a custom field named "Installed" to your components and filling it
# with a value of "NU" (Normally Uninstalled).
# See netlist.getInterestingComponents(), or
#
# 2) blacklisting it in any of the three following lists:


# regular expressions which match component 'Reference' fields of components that
# are to be excluded from the BOM.
excluded_references = [
    # 'TP[0-9]+'              # all test points
    ]


# regular expressions which match component 'Value' fields of components that
# are to be excluded from the BOM.
excluded_values = [
    # 'MOUNTHOLE',
    # 'SCOPETEST',
    # 'MOUNT_HOLE',
    # 'SOLDER_BRIDGE.*'
    ]


# regular expressions which match component 'Footprint' fields of components that
# are to be excluded from the BOM.
excluded_footprints = [
    #'MOUNTHOLE'
    ]

#-----</Configure>---------------------------------------------------------------


class xmlElement():
    """xml element which can represent all nodes of the netlist tree.  It can be
    used to easily generate various output formats by propagating format
    requests to children recursively.
    """
    def __init__(self, name, parent=None):
        self.name = name
        self.attributes = {}
        self.parent = parent
        self.chars = ""
        self.children = []

    def __str__(self):
        """String representation of this netlist element

        """
        return self.name + "[" + self.chars + "]" + " attr_count:" + str(len(self.attributes))

    def formatXML(self, nestLevel=0, amChild=False):
        """Return this element formatted as XML

        Keywords:
        nestLevel -- increases by one for each level of nesting.
        amChild -- If set to True, the start of document is not returned.

        """
        s = ""

        indent = ""
        for i in range(nestLevel):
            indent += "    "

        if not amChild:
            s = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"

        s += indent + "<" + self.name
        for a in self.attributes:
            s += " " + a + "=\"" + self.attributes[a] + "\""

        if (len(self.chars) == 0) and (len(self.children) == 0):
            s += "/>"
        else:
            s += ">" + self.chars

        for c in self.children:
            s += "\n"
            s += c.formatXML(nestLevel+1, True)

        if (len(self.children) > 0):
            s += "\n" + indent

        if (len(self.children) > 0) or (len(self.chars) > 0):
            s += "</" + self.name + ">"

        return s

    def formatHTML(self, amChild=False):
        """Return this element formatted as HTML

        Keywords:
        amChild -- If set to True, the start of document is not returned

        """
        s = ""

        if not amChild:
            s = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
                "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
                <html xmlns="http://www.w3.org/1999/xhtml">
                <head>
                <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
                <title></title>
                </head>
                <body>
                <table>
                """

        s += "<tr><td><b>" + self.name + "</b><br>" + self.chars + "</td><td><ul>"
        for a in self.attributes:
            s += "<li>" + a + " = " + self.attributes[a] + "</li>"

        s += "</ul></td></tr>\n"

        for c in self.children:
            s += c.formatHTML(True)

        if not amChild:
            s += """</table>
                </body>
                </html>"""

        return s

    def addAttribute(self, attr, value):
        """Add an attribute to this element"""
        if type(value) != str: value = value.encode('utf-8')
        self.attributes[attr] = value

    def setAttribute(self, attr, value):
        """Set an attributes value - in fact does the same thing as add
        attribute

        """
        self.attributes[attr] = value

    def setChars(self, chars):
        """Set the characters for this element"""
        self.chars = chars

    def addChars(self, chars):
        """Add characters (textual value) to this element"""
        self.chars += chars

    def addChild(self, child):
        """Add a child element to this element"""
        self.children.append(child)
        return self.children[len(self.children) - 1]

    def getParent(self):
        """Get the parent of this element (Could be None)"""
        return self.parent

    def getChild(self, name):
        """Returns the first child element named 'name'

        Keywords:
        name -- The name of the child element to return"""
        for child in self.children:
            if child.name == name:
                return child
        return None

    def getChildren(self, name=None):
        if name:
            # return _all_ children named "name"
            ret = []
            for child in self.children:
                if child.name == name:
                    ret.append(child)
            return ret
        else:
            return self.children

    def get(self, elemName, attribute="", attrmatch=""):
        """Return the text data for either an attribute or an xmlElement
        """
        if (self.name == elemName):
            if attribute != "":
                try:
                    if attrmatch != "":
                        if self.attributes[attribute] == attrmatch:
                            ret = self.chars
                            if type(ret) != str: ret = ret.encode('utf-8')
                            return ret
                    else:
                        ret = self.attributes[attribute]
                        if type(ret) != str: ret = ret.encode('utf-8')
                        return ret
                except AttributeError:
                    ret = ""
                    if type(ret) != str: ret = ret.encode('utf-8')
                    return ret
            else:
                ret = self.chars
                if type(ret) != str: ret = ret.encode('utf-8')
                return ret

        for child in self.children:
            ret = child.get(elemName, attribute, attrmatch)
            if ret != "":
                if type(ret) != str: ret = ret.encode('utf-8')
                return ret

        ret = ""
        if type(ret) != str: ret = ret.encode('utf-8')
        return ret



class libpart():
    """Class for a library part, aka 'libpart' in the xml netlist file.
    (Components in eeschema are instantiated from library parts.)
    This part class is implemented by wrapping an xmlElement with accessors.
    This xmlElement instance is held in field 'element'.
    """
    def __init__(self, xml_element):
        #
        self.element = xml_element

    #def __str__(self):
        # simply print the xmlElement associated with this part
        #return str(self.element)

    def getLibName(self):
        return self.element.get("libpart", "lib")

    def getPartName(self):
        return self.element.get("libpart", "part")

    def getDescription(self):
        return self.element.get("description")

    def getField(self, name):
        return self.element.get("field", "name", name)

    def getFieldNames(self):
        """Return a list of field names in play for this libpart.
        """
        fieldNames = []
        fields = self.element.getChild('fields')
        if fields:
            for f in fields.getChildren():
                fieldNames.append( f.get('field','name') )
        return fieldNames

    def getPinList(self):
        """Return a list of pins in play for this libpart.
        """
        pinList = []
        pins = self.element.getChild('pins')
        if pins:
            for f in pins.getChildren():
                pinList.append( f )
        return pinList

    def getDatasheet(self):
        return self.getField("Datasheet")

    def getFootprint(self):
        return self.getField("Footprint")

    def getAliases(self):
        """Return a list of aliases or None"""
        aliases = self.element.getChild("aliases")
        if aliases:
            ret = []
            children = aliases.getChildren()
            # grab the text out of each child:
            for child in children:
                ret.append( child.get("alias") )
            return ret
        return None


class comp():
    """Class for a component, aka 'comp' in the xml netlist file.
    This component class is implemented by wrapping an xmlElement instance
    with accessors.  The xmlElement is held in field 'element'.
    """

    def __init__(self, xml_element):
        self.element = xml_element
        self.libpart = None

        # Set to true when this component is included in a component group
        self.grouped = False

    def __eq__(self, other):
        """ Equivalency operator, remember this can be easily overloaded
            2 components are equivalent ( i.e. can be grouped
            if they have same value and same footprint and are both set to be populated

            Override the component equivalence operator must be done before
            loading the netlist, otherwise all components will have the original
            equivalency operator.

            You have to define a comparison module (for instance named myEqu)
            and add the line;
                kicad_netlist_reader.comp.__eq__ = myEqu
            in your bom generator script before calling the netliste reader by something like:
                net = kicad_netlist_reader.netlist(sys.argv[1])
        """
        result = False
        if self.getValue() == other.getValue():
            if self.getFootprint() == other.getFootprint():
                if self.getRef().rstrip(string.digits) == other.getRef().rstrip(string.digits):
                    if self.getDNP() == other.getDNP():
                        result = True
        return result

    def setLibPart(self, part):
        self.libpart = part

    def getLibPart(self):
        return self.libpart

    def getPartName(self):
        return self.element.get("libsource", "part")

    def getLibName(self):
        return self.element.get("libsource", "lib")

    def setValue(self, value):
        """Set the value of this component"""
        v = self.element.getChild("value")
        if v:
            v.setChars(value)

    def getValue(self):
        return self.element.get("value")

    def getField(self, name, aLibraryToo = True):
        """
        Return the value of a field named name. The component is first
        checked for the field, and then the components library part is checked
        for the field. If the field doesn't exist in either, an empty string is
        returned

        Keywords:
        name -- The name of the field to return the value for
        aLibraryToo --  look in the libpart's fields for the same name if not found
                        in component itself
        """

        field = self.element.get("field", "name", name)
        if field == "" and aLibraryToo and self.libpart:
            field = self.libpart.getField(name)
        return field

    def getFieldNames(self):
        """Return a list of field names in play for this component.  Mandatory
        fields are not included, and they are: Value, Footprint, Datasheet, Ref.
        The netlist format only includes fields with non-empty values.  So if a field
        is empty, it will not be present in the returned list.
        """
        fieldNames = []
        fields = self.element.getChild('fields')
        if fields:
            for f in fields.getChildren():
                fieldNames.append( f.get('field','name') )
        return fieldNames

    def getRef(self):
        return self.element.get("comp", "ref")

    '''
    Return true if the component has the DNP property set
    '''
    def getDNP(self):
        for child in self.element.getChildren( "property"):
            try:
                if child.attributes['name'] == 'dnp':
                    return True
            except KeyError:
                continue

        return False

    '''
    Return 'DNP' if the component has the DNP property set
    '''
    def getDNPString(self):
        if self.getDNP():
            return 'DNP'

        return ''

    '''
    Return true if the component has the exclude from BOM property set
    '''
    def getExcludeFromBOM(self):
        for child in self.element.getChildren( "property"):
            try:
                if child.attributes['name'] == 'exclude_from_bom':
                    return True
            except KeyError:
                continue

        return False

    '''
    Return true if the component has the exclude from Board property set
    '''
    def getExcludeFromBoard(self):
        for child in self.element.getChildren( "property"):
            try:
                if child.attributes['name'] == 'exclude_from_board':
                    return True
            except KeyError:
                continue

        return False

    '''
    return the footprint name. if empty and aLibraryToo = True, return the
    footprint name from libary
    '''
    def getFootprint(self, aLibraryToo = True):
        ret = self.element.get("footprint")

        if ret == "" and aLibraryToo and self.libpart:
            ret = self.libpart.getFootprint()

        return ret

    '''
    return the datasheet name. if empty and aLibraryToo = True, return the
    datasheet name from libary
    '''
    def getDatasheet(self, aLibraryToo = True):
        ret = self.element.get("datasheet")
        if ret == "" and aLibraryToo and self.libpart:
            ret = self.libpart.getDatasheet()
        return ret

    def getTimestamp(self):
        """
        Kicad 5 uses tstamp keyword for time stamp (8 digits) as UUID
        Kicad 6 uses tstamps keyword for UUID and a multi unit symbol has more than one UUID
        (UUIDs are separated by spaces)
        """
        ret = self.element.get("tstamp")
        if ret == "":
            ret = self.element.get("tstamps")
        return ret

    def getDescription(self):
        return self.element.get("libsource", "description")

    '''
    return the netname of the pin aPinNum in netlist aNetlist
    if aSkipEmptyNet = True, net having only one pin will return a empty name
    '''
    def getPinNetname(self, aPinNum, aNetlist, aSkipEmptyNet):
        ref = self.getRef()

        for net in aNetlist.nets:
            net_name = net.get( "net", "name" )

            item_cnt = 1
            netitems = net.children

            for node in netitems:
                curr_item_ref = node.get( "node", "ref" )

                if curr_item_ref == ref:
                    curr_pin = node.get( "node", "pin" )

                    if aPinNum == curr_pin:
                        if aSkipEmptyNet:   #ensure at least 2 pins are in net
                            pin_count = 0

                            for curr_node in netitems:
                                pin_count += 1

                                if pin_count > 1:
                                    return net_name

                            return ""
                        else:
                            return net_name

        return "?"


class netlist():
    """ Kicad generic netlist class. Generally loaded from a kicad generic
    netlist file. Includes several helper functions to ease BOM creating
    scripts

    """
    def __init__(self, fname=""):
        """Initialiser for the genericNetlist class

        Keywords:
        fname -- The name of the generic netlist file to open (Optional)

        """
        self.design = None
        self.components = []
        self.libparts = []
        self.libraries = []
        self.nets = []

        # The entire tree is loaded into self.tree
        self.tree = []

        self._curr_element = None

        # component blacklist regexs, made from excluded_* above.
        self.excluded_references = []
        self.excluded_values = []
        self.excluded_footprints = []

        if fname != "":
            self.load(fname)

    def addChars(self, content):
        """Add characters to the current element"""
        self._curr_element.addChars(content)

    def addElement(self, name):
        """Add a new kicad generic element to the list"""
        if self._curr_element is None:
            self.tree = xmlElement(name)
            self._curr_element = self.tree
        else:
            self._curr_element = self._curr_element.addChild(
                xmlElement(name, self._curr_element))

        # If this element is a component, add it to the components list
        if self._curr_element.name == "comp":
            self.components.append(comp(self._curr_element))

        # Assign the design element
        if self._curr_element.name == "design":
            self.design = self._curr_element

        # If this element is a library part, add it to the parts list
        if self._curr_element.name == "libpart":
            self.libparts.append(libpart(self._curr_element))

        # If this element is a net, add it to the nets list
        if self._curr_element.name == "net":
            self.nets.append(self._curr_element)

        # If this element is a library, add it to the libraries list
        if self._curr_element.name == "library":
            self.libraries.append(self._curr_element)

        return self._curr_element

    def endDocument(self):
        """Called when the netlist document has been fully parsed"""
        # When the document is complete, the library parts must be linked to
        # the components as they are separate in the tree so as not to
        # duplicate library part information for every component
        for c in self.components:
            for p in self.libparts:
                if p.getLibName() == c.getLibName():
                    if p.getPartName() == c.getPartName():
                        c.setLibPart(p)
                        break
                    else:
                        aliases = p.getAliases()
                        if aliases and self.aliasMatch( c.getPartName(), aliases ):
                            c.setLibPart(p)
                            break;

            if not c.getLibPart():
                print( 'missing libpart for ref:', c.getRef(), c.getPartName(), c.getLibName() )


    def aliasMatch(self, partName, aliasList):
        for alias in aliasList:
            if partName == alias:
                return True
        return False

    def endElement(self):
        """End the current element and switch to its parent"""
        self._curr_element = self._curr_element.getParent()

    def getDate(self):
        """Return the date + time string generated by the tree creation tool"""
        return self.design.get("date")

    def getSource(self):
        """Return the source string for the design"""
        return self.design.get("source")

    def getTool(self):
        """Return the tool string which was used to create the netlist tree"""
        return self.design.get("tool")

    def getNets(self):
        """Return the nets """
        return self.nets

    def gatherComponentFieldUnion(self, components=None):
        """Gather the complete 'set' of unique component fields, fields found in any component.
        """
        if not components:
            components=self.components

        s = set()
        for c in components:
            s.update( c.getFieldNames() )

        # omit anything matching any regex in excluded_fields
        ret = set()
        for field in s:
            exclude = False
            for rex in excluded_fields:
                if re.match( rex, field ):
                    exclude = True
                    break
            if not exclude:
                ret.add(field)

        return ret       # this is a python 'set'

    def gatherLibPartFieldUnion(self):
        """Gather the complete 'set' of part fields, fields found in any part.
        """
        s = set()
        for p in self.libparts:
            s.update( p.getFieldNames() )

        # omit anything matching any regex in excluded_fields
        ret = set()
        for field in s:
            exclude = False
            for rex in excluded_fields:
                if re.match( rex, field ):
                    exclude = True
                    break
            if not exclude:
                ret.add(field)

        return ret       # this is a python 'set'

    def getInterestingComponents(self, excludeBOM=False, excludeBoard=False, DNP=False):
        """Return a subset of all components, those that should show up in the BOM.
        Omit those that should not, by consulting the blacklists:
        excluded_values, excluded_refs, and excluded_footprints, which hold one
        or more regular expressions.  If any of the regular expressions match
        the corresponding field's value in a component, then the component is excluded.
        """

        # pre-compile all the regex expressions:
        del self.excluded_references[:]
        del self.excluded_values[:]
        del self.excluded_footprints[:]

        for rex in excluded_references:
            self.excluded_references.append( re.compile( rex ) )

        for rex in excluded_values:
            self.excluded_values.append( re.compile( rex ) )

        for rex in excluded_footprints:
            self.excluded_footprints.append( re.compile( rex ) )

        # the subset of components to return, considered as "interesting".
        ret = []

        # run each component thru a series of tests, if it passes all, then add it
        # to the interesting list 'ret'.
        for c in self.components:
            exclude = False
            if not exclude:
                for refs in self.excluded_references:
                    if refs.match(c.getRef()):
                        exclude = True
                        break
            if not exclude:
                for vals in self.excluded_values:
                    if vals.match(c.getValue()):
                        exclude = True
                        break
            if not exclude:
                for mods in self.excluded_footprints:
                    if mods.match(c.getFootprint()):
                        exclude = True
                        break

            if excludeBOM and c.getExcludeFromBOM():
                exclude = True

            if excludeBoard and c.getExcludeFromBoard():
                exclude = True

            if DNP and c.getDNP():
                exclude = True

            if not exclude:
                ret.append(c)

        # The key to sort the components in the BOM
        # This sorts using a natural sorting order (e.g. 100 after 99), and if it wasn't used
        # the normal sort would place 100 before 99 since it only would look at the first digit.
        def sortKey( str ):
            return [ int(t) if t.isdigit() else t.lower()
                    for t in re.split( r'(\d+)', str ) ]

        ret.sort(key=lambda g: sortKey(g.getRef()))

        return ret


    def groupComponents(self, components = None):
        """Return a list of component lists. Components are grouped together
        when the value, library and part identifiers match.

        Keywords:
        components -- is a list of components, typically an interesting subset
        of all components, or None.  If None, then all components are looked at.
        """
        if not components:
            components = self.components

        groups = []

        # Make sure to start off will all components ungrouped to begin with
        for c in components:
            c.grouped = False

        # Group components based on the value, library and part identifiers
        for c in components:
            if c.grouped == False:
                c.grouped = True
                newgroup = []
                newgroup.append(c)

                # Check every other ungrouped component against this component
                # and add to the group as necessary
                for ci in components:
                    if ci.grouped == False and ci == c:
                        newgroup.append(ci)
                        ci.grouped = True

                # Add the new component group to the groups list
                groups.append(newgroup)

        # The key to sort the components in the BOM
        # This sorts using a natural sorting order (e.g. 100 after 99), and if it wasn't used
        # the normal sort would place 100 before 99 since it only would look at the first digit.
        def sortKey( str ):
            return [ int(t) if t.isdigit() else t.lower()
                    for t in re.split( r'(\d+)', str ) ]

        for g in groups:
            #g = sorted(g, key=lambda g: sortKey(g.getRef()))
            g.sort(key=lambda g: sortKey(g.getRef()))

        # Finally, sort the groups to order the references alphabetically
        groups.sort(key=lambda group: sortKey(group[0].getRef()))

        return groups

    def getGroupField(self, group, field):
        """Return the whatever is known about the given field by consulting each
        component in the group.  If any of them know something about the property/field,
        then return that first non-blank value.
        """
        for c in group:
            ret = c.getField(field, False)
            if ret != '':
                return ret

        libpart = group[0].getLibPart()
        if not libpart:
            return ''

        return libpart.getField(field)

    def getGroupFootprint(self, group):
        """Return the whatever is known about the Footprint by consulting each
        component in the group.  If any of them know something about the Footprint,
        then return that first non-blank value.
        """
        for c in group:
            ret = c.getFootprint()
            if ret != "":
                return ret
        return group[0].getLibPart().getFootprint()

    def getGroupDatasheet(self, group):
        """Return the whatever is known about the Datasheet by consulting each
        component in the group.  If any of them know something about the Datasheet,
        then return that first non-blank value.
        """
        for c in group:
            ret = c.getDatasheet()
            if ret != "":
                return ret

        if len(group) > 0:
            return group[0].getLibPart().getDatasheet()
        else:
            print("NULL!")
        return ''

    def formatXML(self):
        """Return the whole netlist formatted in XML"""
        return self.tree.formatXML()

    def formatHTML(self):
        """Return the whole netlist formatted in HTML"""
        return self.tree.formatHTML()

    def load(self, fname):
        """Load a kicad generic netlist

        Keywords:
        fname -- The name of the generic netlist file to open

        """
        try:
            self._reader = sax.make_parser()
            self._reader.setContentHandler(_gNetReader(self))
            self._reader.parse(fname)
        except IOError as e:
            print( __file__, ":", e, file=sys.stderr )
            sys.exit(-1)



class _gNetReader(sax.handler.ContentHandler):
    """SAX kicad generic netlist content handler - passes most of the work back
    to the 'netlist' class which builds a complete tree in RAM for the design

    """
    def __init__(self, aParent):
        self.parent = aParent

    def startElement(self, name, attrs):
        """Start of a new XML element event"""
        element = self.parent.addElement(name)

        for name in attrs.getNames():
            element.addAttribute(name, attrs.getValue(name))

    def endElement(self, name):
        self.parent.endElement()

    def characters(self, content):
        # Ignore erroneous white space - ignoreableWhitespace does not get rid
        # of the need for this!
        if not content.isspace():
            self.parent.addChars(content)

    def endDocument(self):
        """End of the XML document event"""
        self.parent.endDocument()



#
# Example python script to generate a BOM from a KiCad generic netlist
#
# Example: Sorted and Grouped CSV BOM
#

"""
    @package
    Output: CSV (comma-separated)
    Grouped By: Value, Footprint, DNP
    Sorted By: Ref
    Fields: Ref, Qnty, Value, Cmp name, Footprint, Description, Vendor, DNP

    Command line:
    python "pathToFile/bom_csv_grouped_by_value_with_fp.py" "%I" "%O.csv"
"""

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import kicad_utils
import csv
import sys

# A helper function to filter/convert a string read in netlist
#currently: do nothing
def fromNetlistText( aText ):
    return aText

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = kicad_utils.open_file_writeUTF8(sys.argv[2], 'w')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

# Create a new csv writer object to use as the output formatter
out = csv.writer(f, lineterminator='\n', delimiter=',', quotechar='\"', quoting=csv.QUOTE_ALL)

# Output a set of rows for a header providing general information
out.writerow(['Source:', net.getSource()])
out.writerow(['Date:', net.getDate()])
out.writerow(['Tool:', net.getTool()])
out.writerow( ['Generator:', sys.argv[0]] )
out.writerow(['Component Count:', len(net.components)])
out.writerow(['Ref', 'Qnty', 'Value', 'Cmp name', 'Footprint', 'Description', 'Vendor', 'DNP'])

# subset the components to those wanted in the BOM, controlled
# by <configure> block in kicad_netlist_reader.py
components = net.getInterestingComponents( excludeBOM=True )

# Get all of the components in groups of matching parts + values
# (see ky_generic_netlist_reader.py)
grouped = net.groupComponents(components)

# Output all of the component information
for group in grouped:
    refs = ""
    refs_l = []

    # Add the reference of every component in the group and keep a reference
    # to the component so that the other data can be filled in once per group
    for component in group:
        refs_l.append( fromNetlistText( component.getRef() ) )
        c = component

    refs = ", ".join(refs_l)

    # Fill in the component groups common data
    out.writerow([refs, len(group),
        fromNetlistText( c.getValue() ),
        fromNetlistText( c.getPartName() ),
        fromNetlistText( c.getFootprint() ),
        fromNetlistText( c.getDescription() ),
        fromNetlistText( c.getField("Vendor") ),
        c.getDNPString()])

if not canOpenFile:
    sys.exit(1)



#
# Example python script to generate a BOM from a KiCad generic netlist
#
# Example: Ungrouped (One component per row) CSV output
#

"""
    @package
    Output: CSV (comma-separated)
    Grouped By: ungrouped, one component per line
    Sorted By: Ref
    Fields: Ref, Value, Part, Footprint, Datasheet, Manufacturer, Vendor, DNP

    Command line:
    python "pathToFile/bom_csv_sorted_by_ref.py" "%I" "%O.csv"
"""

from __future__ import print_function

# Import the KiCad python helper module
import kicad_netlist_reader
import kicad_utils
import csv
import sys

# A helper function to filter/convert a string read in netlist
#currently: do nothing
def fromNetlistText( aText ):
    return aText

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = kicad_utils.open_file_writeUTF8(sys.argv[2], 'w')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print( __file__, ":", e, sys.stderr )
    f = sys.stdout
    canOpenFile = False

# Create a new csv writer object to use as the output formatter
out = csv.writer(f, lineterminator='\n', delimiter=',', quotechar="\"", quoting=csv.QUOTE_ALL)

# override csv.writer's writerow() to support utf8 encoding:
def writerow( acsvwriter, columns ):
    utf8row = []
    for col in columns:
        utf8row.append( fromNetlistText( str(col) ) )
    acsvwriter.writerow( utf8row )

components = net.getInterestingComponents( excludeBOM=True )

# Output a field delimited header line
writerow( out, ['Source:', net.getSource()] )
writerow( out, ['Date:', net.getDate()] )
writerow( out, ['Tool:', net.getTool()] )
writerow( out, ['Component Count:', len(components)] )
writerow( out, ['Ref', 'Value', 'Footprint', 'Datasheet', 'Manufacturer', 'Vendor', 'DNP'] )

# Output all of the component information (One component per row)
for c in components:
    writerow( out, [c.getRef(), c.getValue(), c.getFootprint(), c.getDatasheet(),
        c.getField("Manufacturer"), c.getField("Vendor"), c.getDNPString()])


if not canOpenFile:
    sys.exit(1)